# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Utility script for comparing the internals of two Java class files
for differences in structure and data. Has options to specify changes
which may be immaterial or unimportant, such as re-ordering of the
constant pool, line number changes (either absolute or relative),
added fields or methods, deprecation changes, etc.

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL
"""


import sys

from abc import ABCMeta
from argparse import ArgumentParser, Action
from six import add_metaclass

from . import unpack_classfile
from .change import GenericChange, SuperChange
from .change import Addition, Removal
from .change import yield_sorted_by_type
from .opcodes import get_opname_by_code, has_const_arg
from .report import quick_report, Reporter
from .report import JSONReportFormat, TextReportFormat
from .report import add_general_report_optgroup
from .report import add_json_report_optgroup, add_html_report_optgroup


__all__ = (
    "JavaClassChange",
    "ClassInfoChange",
    "ClassAnnotationsChange",
    "ClassInvisibleAnnotationsChange",
    "ClassConstantPoolChange",
    "ClassFieldsChange", "ClassMethodsChange",

    "ClassMembersChange",
    "MemberSuperChange", "MemberAdded", "MemberRemoved",

    "FieldChange", "FieldAdded", "FieldRemoved",
    "FieldNameChange", "FieldTypeChange",
    "FieldSignatureChange", "FieldAnnotationsChange",
    "FieldInvisibleAnnotationsChange",
    "FieldAccessflagsChange", "FieldConstvalueChange",
    "FieldDeprecationChange",

    "MethodChange", "MethodAdded", "MethodRemoved",
    "MethodNameChange", "MethodTypeChange",
    "MethodSignatureChange", "MethodAnnotationsChange",
    "MethodInvisibleAnnotationsChange",
    "MethodParametersChange", "MethodAccessflagsChange",
    "MethodExceptionsChange",
    "MethodAbstractChange", "MethodCodeChange",
    "MethodDeprecationChange",

    "CodeAbsoluteLinesChange",
    "CodeRelativeLinesChange",
    "CodeStackChange",
    "CodeLocalsChange",
    "CodeExceptionChange",
    "CodeConstantsChange",
    "CodeBodyChange",

    "JavaClassReport",

    "pretty_merge_constants", "merge_code",

    "cli", "main",
    "cli_classes_diff",
    "add_classdiff_optgroup", "default_classdiff_options",
    "add_general_optgroup", )


class ClassNameChange(GenericChange):

    label = "Class name"


    def fn_data(self, c):
        return c.get_this()


    def fn_pretty(self, c):
        return c.pretty_this()


class ClassVersionChange(GenericChange):

    label = "Java class verison"


    def fn_data(self, c):
        return c.version


    def is_ignored(self, o):
        lver = self.ldata.version
        rver = self.rdata.version

        return ((o.ignore_version_up and lver < rver) or
                (o.ignore_version_down and lver > rver))


class ClassPlatformChange(GenericChange):

    label = "Java platform"


    def fn_data(self, c):
        return c.get_platform()


    def is_ignored(self, o):
        lver = self.ldata.version
        rver = self.rdata.version

        return ((o.ignore_version_up and lver < rver) or
                (o.ignore_version_down and lver > rver))


class ClassSuperclassChange(GenericChange):

    label = "Superclass"


    def fn_data(self, c):
        return c.get_super()


    def fn_pretty(self, c):
        return c.pretty_super()


class ClassInterfacesChange(GenericChange):

    label = "Interfaces"


    def fn_data(self, c):
        return set(c.get_interfaces())


    def fn_pretty(self, c):
        return tuple(c.pretty_interfaces())


class ClassAccessflagsChange(GenericChange):

    label = "Access flags"


    def fn_data(self, c):
        return c.access_flags


    def fn_pretty(self, c):
        return tuple(c.pretty_access_flags())


class ClassDeprecationChange(GenericChange):

    label = "Deprecation"


    def fn_data(self, c):
        return c.is_deprecated()


    def is_ignored(self, o):
        return o.ignore_deprecated


class ClassSignatureChange(GenericChange):

    label = "Generics Signature"


    def fn_data(self, c):
        return c.get_signature()


    def fn_pretty(self, c):
        return c.pretty_signature()


@add_metaclass(ABCMeta)
class AnnotationsChange(GenericChange):

    label = "Runtime annotations"


    def fn_data(self, c):
        return c.get_annotations() or tuple()


    def fn_pretty(self, c):
        annos = self.fn_data(c)
        return [anno.pretty_annotation() for anno in annos]


@add_metaclass(ABCMeta)
class InvisibleAnnotationsChange(AnnotationsChange):

    label = "Runtime Invisible annotations"


    def fn_data(self, c):
        return c.get_invisible_annotations() or tuple()


class ClassAnnotationsChange(AnnotationsChange):

    label = "Class runtime annotations"


class ClassInvisibleAnnotationsChange(InvisibleAnnotationsChange):

    label = "Class runtime invisible annotations"


class ClassInfoChange(SuperChange):

    label = "Class information"


    change_types = (ClassNameChange,
                    ClassVersionChange,
                    ClassPlatformChange,
                    ClassSuperclassChange,
                    ClassInterfacesChange,
                    ClassAccessflagsChange,
                    ClassDeprecationChange,
                    ClassSignatureChange)


@add_metaclass(ABCMeta)
class MemberSuperChange(SuperChange):
    """
    basis for FieldChange and MethodChange
    """

    label = "Member"


    def get_description(self):
        return "%s: %s" % (self.label, self.ldata.pretty_descriptor())


@add_metaclass(ABCMeta)
class MemberAdded(Addition):
    """
    basis for FieldAdded and MethodAdded
    """

    label = "Member added"


    def get_description(self):
        return "%s: %s" % (self.label, self.rdata.pretty_descriptor())


@add_metaclass(ABCMeta)
class MemberRemoved(Removal):
    """
    basis for FieldChange and MethodChange
    """

    label = "Member removed"


    def get_description(self):
        return "%s: %s" % (self.label, self.ldata.pretty_descriptor())


@add_metaclass(ABCMeta)
class ClassMembersChange(SuperChange):
    """
    basis for ClassFieldsChange and ClassMethodsChange
    """

    label = "Members"

    member_added = MemberAdded
    member_removed = MemberRemoved
    member_changed = MemberSuperChange


    def collect_impl(self):
        li = {}

        for member in self.ldata:
            li[member.get_identifier()] = member

        for member in self.rdata:
            key = member.get_identifier()
            lf = li.get(key, None)

            if lf:
                del li[key]
                yield self.member_changed(lf, member)
            else:
                yield self.member_added(None, member)

        for member in li.values():
            yield self.member_removed(member, None)


class CodeAbsoluteLinesChange(GenericChange):

    label = "Absolute line numbers"


    def fn_data(self, c):
        return (c and c.get_linenumbertable()) or tuple()


    def is_ignored(self, options):
        return options.ignore_absolute_lines



class CodeRelativeLinesChange(GenericChange):

    label = "Relative line numbers"


    def fn_data(self, c):
        return (c and c.get_relativelinenumbertable()) or tuple()


    def is_ignored(self, options):
        return options.ignore_relative_lines



class CodeStackChange(GenericChange):

    label = "Stack size"


    def fn_data(self, c):
        return (c and c.max_stack) or 0



class CodeLocalsChange(GenericChange):

    label = "Locals"


    def fn_data(self, c):
        return (c and c.max_locals) or 0


class CodeExceptionChange(GenericChange):

    label = "Exception table"


    def fn_data(self, c):
        return (c and c.exceptions) or tuple()


    def fn_pretty(self, c):
        a = list()
        for e in self.fn_data(c):
            p = (e.start_pc, e.end_pc,
                 e.handler_pc, e.pretty_catch_type())
            a.append(p)
        return repr(a)


class CodeConstantsChange(GenericChange):
    """
    This is a test to find instances where the individual opcodes and
    arguments for a method's code may all be identical except that ops
    which load from the constant pool may use a different index. We
    will deref the constant index for both sides, and if all of the
    constant values match then we can consider the code to be equal.

    The purpose of such a check is to find other-wise meaningless
    constant pool reordering. If all uses of the pool result in the
    same values, we don't really care if the pool is in a different
    order between the old and new versions of a class.
    """

    label = "Code constants"


    def __init__(self, lcode, rcode):
        super(CodeConstantsChange, self).__init__(lcode, rcode)
        self.offsets = None


    def fn_pretty(self, c):
        if not self.offsets:
            return None

        pr = list()

        for offset, code, args in c.disassemble():
            if offset in self.offsets and has_const_arg(code):
                name = get_opname_by_code(code)
                data = c.cpool.pretty_deref_const(args[0])
                pr.append((offset, name, data))

        return pr


    def check_impl(self):
        left = self.ldata
        right = self.rdata
        offsets = list()

        if left is None or right is None:
            return True, None

        if len(left.code) != len(right.code):
            # code body change, can't determine constants
            return True, None

        for l, r in zip(left.disassemble(), right.disassemble()):
            if not ((l[0] == r[0]) and (l[1] == r[1])):
                # code body change, can't determine constants
                return True, None

            largs = l[2]
            rargs = r[2]

            if has_const_arg(l[1]):
                largs, rargs = list(largs), list(rargs)
                largs[0] = left.cpool.deref_const(largs[0])
                rargs[0] = right.cpool.deref_const(rargs[0])

            if largs != rargs:
                offsets.append(l[0])

        self.offsets = offsets
        return bool(self.offsets), None


class CodeBodyChange(GenericChange):
    """
    The length or the opcodes or the arguments of the opcodes has
    changed, signalling that the method body is different
    """

    label = "Code body"


    def fn_data(self, c):
        return (c and c.disassemble()) or tuple()


    def fn_pretty(self, c):
        pr = list()
        for offset, code, args in self.fn_data(c):
            name = get_opname_by_code(code)
            pr.append((offset, name, args))

        return pr


    def check_impl(self):
        left = self.ldata
        right = self.rdata

        if left is None or right is None:
            return True, None

        if len(left.code) != len(right.code):
            desc = "Code length changed from %r to %r" % \
                   (len(left.code), len(right.code))
            return True, desc

        for l, r in zip(left.disassemble(), right.disassemble()):
            if not ((l[0] == r[0]) and (l[1] == r[1])):
                return True, None

        return False, None


class MethodNameChange(GenericChange):

    label = "Method name"


    def fn_data(self, c):
        return c.get_name()


class MethodTypeChange(GenericChange):

    label = "Method type"


    def fn_data(self, c):
        return c.get_type_descriptor()


    def fn_pretty(self, c):
        return c.pretty_type()


class MethodSignatureChange(GenericChange):

    label = "Method generic signature"


    def fn_data(self, c):
        return c.get_signature()


    def fn_pretty(self, c):
        return c.pretty_signature()


class MethodParametersChange(GenericChange):

    label = "Method parameters"


    def fn_data(self, c):
        return c.get_arg_type_descriptors()


    def fn_pretty(self, c):
        return tuple(c.pretty_arg_types())


class MethodAccessflagsChange(GenericChange):

    label = "Method accessflags"


    def fn_data(self, c):
        return c.access_flags


    def fn_pretty(self, c):
        return tuple(c.pretty_access_flags())


class MethodAbstractChange(GenericChange):

    label = "Method abstract"


    def fn_data(self, c):
        return not c.get_code()


    def fn_pretty_desc(self, c):
        if self.fn_data(c):
            return "Method is abstract"
        else:
            return "Method is concrete"


class MethodExceptionsChange(GenericChange):

    label = "Method exceptions"


    def fn_data(self, c):
        return c.get_exceptions()


    def fn_pretty(self, c):
        return tuple(c.pretty_exceptions())


class MethodCodeChange(SuperChange):

    label = "Method Code"


    change_types = (CodeAbsoluteLinesChange,
                    CodeRelativeLinesChange,
                    CodeStackChange,
                    CodeLocalsChange,
                    CodeExceptionChange,
                    CodeConstantsChange,
                    CodeBodyChange)


    def __init__(self, l, r):
        super(MethodCodeChange, self).__init__(l.get_code(),
                                               r.get_code())


    def collect_impl(self):
        # if both sides are abstract, don't bother collecting subchanges

        if self.ldata is self.rdata is None:
            return tuple()
        else:
            return super(MethodCodeChange, self).collect_impl()


    def check_impl(self):
        # if one side has gone abstract, don't bother trying to check
        # any deeper, they're different unless they're both abstract.

        if None in (self.ldata, self.rdata):
            return (self.ldata != self.rdata), None
        else:
            return super(MethodCodeChange, self).check_impl()


class MethodDeprecationChange(GenericChange):

    label = "Method deprecation"


    def fn_data(self, c):
        return c.is_deprecated()


    def is_ignored(self, o):
        return o.ignore_deprecated


class MethodAnnotationsChange(AnnotationsChange):

    label = "Method runtime annotations"


class MethodInvisibleAnnotationsChange(InvisibleAnnotationsChange):

    label = "Method runtime invisible annotations"


class MethodChange(MemberSuperChange):

    label = "Method"


    change_types = (MethodNameChange,
                    MethodTypeChange,
                    MethodSignatureChange,
                    MethodAnnotationsChange,
                    MethodInvisibleAnnotationsChange,
                    MethodParametersChange,
                    MethodAccessflagsChange,
                    MethodExceptionsChange,
                    MethodAbstractChange,
                    MethodDeprecationChange,
                    MethodCodeChange)


class FieldNameChange(GenericChange):

    label = "Field name"


    def fn_data(self, c):
        return c.get_name()


class FieldTypeChange(GenericChange):

    label = "Field type"


    def fn_data(self, c):
        return c.get_descriptor()


    def fn_pretty(self, c):
        return c.pretty_type()


class FieldSignatureChange(GenericChange):

    label = "Field Generic Signature"


    def fn_data(self, c):
        return c.get_signature()


    def fn_pretty(self, c):
        return c.pretty_signature()


class FieldAccessflagsChange(GenericChange):

    label = "Field accessflags"


    def fn_data(self, c):
        return c.access_flags


    def fn_pretty(self, c):
        return tuple(c.pretty_access_flags())


    def fn_pretty_desc(self, c):
        return ",".join(c.pretty_access_flags())


class FieldConstvalueChange(GenericChange):

    label = "Field constvalue"


    def fn_data(self, c):
        return c.deref_constantvalue()


    def fn_pretty(self, c):
        return repr(c.deref_constantvalue())


class FieldDeprecationChange(GenericChange):

    label = "Field deprecation"


    def fn_data(self, c):
        return c.is_deprecated()


    def is_ignored(self, o):
        return o.ignore_deprecated


class FieldAnnotationsChange(AnnotationsChange):

    label = "Field runtime annotations"


class FieldInvisibleAnnotationsChange(InvisibleAnnotationsChange):

    label = "Field runtime invisible annotations"


class FieldChange(MemberSuperChange):

    label = "Field"


    change_types = (FieldNameChange,
                    FieldTypeChange,
                    FieldSignatureChange,
                    FieldAnnotationsChange,
                    FieldInvisibleAnnotationsChange,
                    FieldAccessflagsChange,
                    FieldConstvalueChange,
                    FieldDeprecationChange)


class FieldAdded(MemberAdded):

    label = "Field added"


class FieldRemoved(MemberRemoved):

    label = "Field removed"


class ClassFieldsChange(ClassMembersChange):

    label = "Fields"


    member_added = FieldAdded
    member_removed = FieldRemoved
    member_changed = FieldChange


    @yield_sorted_by_type(FieldAdded, FieldRemoved, FieldChange)
    def collect_impl(self):
        return super(ClassFieldsChange, self).collect_impl()


    def __init__(self, lclass, rclass):
        super(ClassFieldsChange, self).__init__(lclass.fields,
                                                rclass.fields)


class MethodAdded(MemberAdded):

    label = "Method added"


class MethodRemoved(MemberRemoved):

    label = "Method removed"


class ClassMethodsChange(ClassMembersChange):

    label = "Methods"


    member_added = MethodAdded
    member_removed = MethodRemoved
    member_changed = MethodChange


    @yield_sorted_by_type(MethodAdded, MethodRemoved, MethodChange)
    def collect_impl(self):
        return super(ClassMethodsChange, self).collect_impl()


    def __init__(self, lclass, rclass):
        super(ClassMethodsChange, self).__init__(lclass.methods,
                                                 rclass.methods)


class ClassConstantPoolChange(GenericChange):

    label = "Constant pool"


    def fn_data(self, c):
        return c.cpool


    def fn_pretty(self, c):
        return tuple(c.cpool.pretty_constants())


    def is_ignored(self, options):
        return options.ignore_pool


    def get_description(self):
        return self.label + ((" unaltered", " altered")[self.is_change()])


class JavaClassChange(SuperChange):

    label = "Java Class"


    change_types = (ClassInfoChange,
                    ClassAnnotationsChange,
                    ClassInvisibleAnnotationsChange,
                    ClassConstantPoolChange,
                    ClassFieldsChange,
                    ClassMethodsChange)


    def get_description(self):
        return "%s %s" % (self.label, self.ldata.pretty_this())


class JavaClassReport(JavaClassChange):
    """
    a JavaClassChange with the side-effect of writing reports
    """


    def __init__(self, l, r, reporter):
        super(JavaClassReport, self).__init__(l, r)
        self.reporter = reporter


    def check(self):
        super(JavaClassReport, self).check()
        self.reporter.run(self)


# ---- Utility functions ----
#


def pretty_merge_constants(left_cpool, right_cpool):
    """
    sequence of tuples containing (index, left type, left pretty
    value, right type, right pretty value). If the constant pools are
    of inequal length, a value of None will be set in place of the
    type and pretty value for indexes past its end
    """

    lsize = len(left_cpool.consts)
    rsize = len(right_cpool.consts)

    index = 1
    for index in range(1, min(lsize, rsize)):
        lt, lv = left_cpool.pretty_const(index)
        rt, rv = right_cpool.pretty_const(index)
        yield (index, lt, lv, rt, rv)

    if lsize > rsize:
        for index in range(index, lsize):
            lt, lv = left_cpool.pretty_const(index)
            yield (index, lt, lv, None, None)

    elif rsize > lsize:
        for index in range(index, rsize):
            rt, rv = right_cpool.pretty_const(index)
            yield (index, None, None, rt, rv)


def merge_code(left_code, right_code):
    """
    { relative_line:
      ((left_abs_line, ((offset, op, args), ...)),
       (right_abs_line, ((offset, op, args), ...))),
      ... }
    """

    data = dict()

    code_lines = (left_code and left_code.iter_code_by_lines()) or tuple()
    for abs_line, rel_line, dis in code_lines:
        assert(rel_line is not None)
        data[rel_line] = [(abs_line, dis), None]

    code_lines = (right_code and right_code.iter_code_by_lines()) or tuple()
    for abs_line, rel_line, dis in code_lines:
        assert(rel_line is not None)
        found = data.get(rel_line, None)
        if found is None:
            found = [None, (abs_line, dis)]
            data[rel_line] = found
        else:
            found[1] = (abs_line, dis)

    return data


# ---- Begin classdiff CLI code ----
#


def cli_classes_diff(options, left, right):
    reports = getattr(options, "reports", tuple())
    if reports:
        rdir = options.report_dir or "./"

        rpt = Reporter(rdir, "JavaClassReport", options)
        rpt.add_formats_by_name(reports)

        delta = JavaClassReport(left, right, rpt)

    else:
        delta = JavaClassChange(left, right)

    delta.check()

    if not options.silent:
        if options.json:
            quick_report(JSONReportFormat, delta, options)
        else:
            quick_report(TextReportFormat, delta, options)

    if (not delta.is_change()) or delta.is_ignored(options):
        return 0
    else:
        return 1


def cli(options):
    left = unpack_classfile(options.classfile[0])
    right = unpack_classfile(options.classfile[1])

    return cli_classes_diff(options, left, right)


def add_classdiff_optgroup(parser):
    """
    option group specific to class checking
    """

    g = parser.add_argument_group("Class Checking Options")

    g.add_argument("--ignore-version-up", action="store_true",
                   default=False)
    g.add_argument("--ignore-version-down", action="store_true", default=False)
    g.add_argument("--ignore-platform-up", action="store_true", default=False)
    g.add_argument("--ignore-platform-down", action="store_true",
                   default=False)
    g.add_argument("--ignore-absolute-lines", action="store_true",
                   default=False)
    g.add_argument("--ignore-relative-lines", action="store_true",
                   default=False)
    g.add_argument("--ignore-deprecated", action="store_true", default=False)
    g.add_argument("--ignore-added", action="store_true", default=False)
    g.add_argument("--ignore-pool", action="store_true", default=False)

    g.add_argument("--ignore-lines", nargs=0,
                   help="ignore relative and absolute line-number changes",
                   action=_opt_cb_ign_lines)

    g.add_argument("--ignore-platform", nargs=0,
                   help="ignore platform changes",
                   action=_opt_cb_ign_platform)

    g.add_argument("--ignore-version", nargs=0,
                   help="ignore version changes",
                   action=_opt_cb_ign_version)


class _opt_cb_ignore(Action):
    """
    handle the --ignore option, which trigges other options
    """

    def __call__(self, parser, options, values, option_string=None):

        if not values:
            return

        ignore = getattr(options, "ignore", None)
        if ignore is None:
            options.ignore = ignore = list()

        ign = (i.strip() for i in values.split(","))
        ign = (i for i in ign if i)
        for i in ign:
            ignore.append(i)
            setattr(options, "ignore_" + i.replace("-", "_"), True)


class _opt_cb_ign_lines(Action):
    """
    handle the --ignore-lines option
    """

    def __call__(self, parser, options, values, option_string=None):
        options.ignore_lines = True
        options.ignore_absolute_lines = True
        options.ignore_relative_lines = True


class _opt_cb_ign_version(Action):
    """
    handle the --ignore-version option
    """

    def __call__(self, parser, options, values, option_string=None):
        options.ignore_version = True
        options.ignore_version_up = True
        options.ignore_version_down = True


class _opt_cb_ign_platform(Action):
    """
    handle the --ignore-platform option
    """

    def __call__(self, parser, options, values, option_string=None):
        options.ignore_platform = True
        options.ignore_platform_up = True
        options.ignore_platform_down = True


class _opt_cb_verbose(Action):
    """
    handle the --verbose option
    """

    def __call__(self, parser, options, values, option_string=None):
        options.verbose = True
        options.show_unchanged = True
        options.show_ignored = True


def add_general_optgroup(parser):
    """
    option group for general-use features of all javatool CLIs
    """

    g = parser.add_argument_group("General Options")

    g.add_argument("-q", "--quiet", dest="silent",
                   action="store_true", default=False)

    g.add_argument("-v", "--verbose", nargs=0, action=_opt_cb_verbose)

    g.add_argument("-o", "--output", dest="output", default=None)

    g.add_argument("-j", "--json", dest="json",
                   action="store_true", default=False)

    g.add_argument("--show-ignored", action="store_true", default=False)
    g.add_argument("--show-unchanged", action="store_true", default=False)

    g.add_argument("--ignore", action=_opt_cb_ignore,
                   help="comma-separated list of ignores")


def create_optparser(progname=None):
    """
    an OptionParser instance with the appropriate options and groups
    for the classdiff utility
    """

    parser = ArgumentParser(prog=progname)
    parser.add_argument("classfile", nargs=2,
                        help="class files to compare")
    add_general_optgroup(parser)
    add_classdiff_optgroup(parser)

    add_general_report_optgroup(parser)
    add_json_report_optgroup(parser)
    add_html_report_optgroup(parser)

    return parser


def default_classdiff_options(updates=None):
    """
    generate an options object with the appropriate default values in
    place for API usage of classdiff features. overrides is an
    optional dictionary which will be used to update fields on the
    options object.
    """

    parser = create_optparser()
    options, _args = parser.parse_args(list())

    if updates:
        # pylint: disable=W0212
        options._update_careful(updates)

    return options


def main(args=sys.argv):
    """
    Main entry point for the classdiff CLI
    """

    parser = create_optparser(args[0])
    return cli(parser.parse_args(args[1:]))


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Module and utility for fetching information out of a JAR file, and
printing it out.

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL
"""


from __future__ import print_function


import sys

from argparse import ArgumentParser
from json import dump

from . import unpack_class
from .classinfo import cli_print_classinfo, add_classinfo_optgroup
from .dirutils import fnmatches
from .ziputils import open_zip_entry, zip_file, zip_entry_rollup
from .manifest import Manifest


__all__ = (
    "JAR_PATTERNS", "JarInfo",
    "main", "cli", "add_jarinfo_optgroup",
    "cli_jar_classes", "cli_jar_manifest_info",
    "cli_jar_provides", "cli_jar_requires",
    "cli_jar_zip_info", "cli_jarinfo",
    "cli_jarinfo_json",
)


# for reference by other modules
JAR_PATTERNS = (
    "*.ear",
    "*.jar",
    "*.rar",
    "*.sar",
    "*.war",
)


REQ_BY_CLASS = "class.requires"
PROV_BY_CLASS = "class.provides"


class JarInfo(object):

    def __init__(self, filename=None, zipfile=None):
        if not (filename or zipfile):
            raise TypeError("one of pathname or zipinfo must be specified")

        self.filename = filename
        self.zipfile = zipfile

        self._requires = None
        self._provides = None


    def __del__(self):
        self.close()


    def __enter__(self):
        return self


    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return exc_type is None


    def open(self, entry, mode='r'):
        return open_zip_entry(self.get_zipfile(), entry, mode)


    def _collect_requires_provides(self):
        req = {}
        prov = {}

        # we need to collect private provides in order to satisfy deps
        # for things like anonymous inner classes, which have access
        # to private members of their parents. This is only used to
        # filter out false-positive requirements.
        p = set()

        for entry in self.get_classes():
            ci = self.get_classinfo(entry)
            for sym in ci.get_requires():
                req.setdefault(sym, list()).append((REQ_BY_CLASS, entry))
            for sym in ci.get_provides(private=False):
                prov.setdefault(sym, list()).append((PROV_BY_CLASS, entry))
            for sym in ci.get_provides(private=True):
                p.add(sym)

        req = dict((k, v) for k, v in req.items() if k not in p)

        self._requires = req
        self._provides = prov


    def get_requires(self, ignored=tuple()):
        if self._requires is None:
            self._collect_requires_provides()

        d = self._requires
        if ignored:
            d = dict((k, v) for k, v in d.items()
                     if not fnmatches(k, *ignored))
        return d


    def get_provides(self, ignored=tuple()):
        if self._provides is None:
            self._collect_requires_provides()

        d = self._provides
        if ignored:
            d = dict((k, v) for k, v in d.items()
                     if not fnmatches(k, *ignored))
        return d


    def get_classes(self):
        """
        sequence of .class files in the underlying zip
        """

        for n in self.get_zipfile().namelist():
            if fnmatches(n, "*.class"):
                yield n


    def get_classinfo(self, entry):
        """
        fetch a class entry as a JavaClassInfo instance
        """

        with self.open(entry) as cfd:
            return unpack_class(cfd)


    def get_zipfile(self):
        if self.zipfile is None:
            self.zipfile = zip_file(self.filename)
        return self.zipfile


    def get_manifest(self):
        mf = Manifest()
        mf.load_from_jar(self.filename)
        return mf


    def close(self):
        if self.zipfile:
            self.zipfile.close()
            self.zipfile = None


def cli_jar_manifest_info(jarinfo):
    mf = jarinfo.get_manifest()

    if not mf:
        print("Manifest not present.")
        print()
        return

    print("Manifest main section:")
    for k, v in sorted(mf.items()):
        print("  %s: %s" % (k, v))

    for _name, sect in sorted(mf.sub_sections.items()):
        print()
        print("Manifest sub-section:")
        for k, v in sorted(sect.items()):
            print("  %s: %s" % (k, v))

    print()


def cli_jar_zip_info(jarinfo):
    zipfile = jarinfo.get_zipfile()

    files, dirs, comp, uncomp = zip_entry_rollup(zipfile)
    prcnt = (float(comp) / float(uncomp)) * 100

    print("Contains %i files, %i directories" % (files, dirs))
    print("Uncompressed size is %i" % uncomp)
    print("Compressed size is %i (%0.1f%%)" % (comp, prcnt))
    print()


def cli_jar_classes(options, jarinfo):
    for entry in jarinfo.get_classes():
        ci = jarinfo.get_classinfo(entry)
        print("Entry: ", entry)
        cli_print_classinfo(options, ci)
        print()


def cli_jar_provides(options, jarinfo):
    print("jar provides:")

    for provided in sorted(jarinfo.get_provides().keys()):
        if not fnmatches(provided, *options.api_ignore):
            print(" ", provided)
    print()


def cli_jar_requires(options, jarinfo):
    print("jar requires:")

    for required in sorted(jarinfo.get_requires().keys()):
        if not fnmatches(required, *options.api_ignore):
            print(" ", required)
    print()


def cli_jarinfo(options, info):
    if options.zip:
        cli_jar_zip_info(info)

    if options.manifest:
        cli_jar_manifest_info(info)

    if options.jar_provides:
        cli_jar_provides(options, info)

    if options.jar_requires:
        cli_jar_requires(options, info)

    if options.jar_classes:
        cli_jar_classes(options, info)


def cli_jarinfo_json(options, info):
    data = {}

    if options.jar_provides:
        data["jar.provides"] = info.get_provides(options.api_ignore)

    if options.jar_requires:
        data["jar.requires"] = info.get_requires(options.api_ignore)

    if options.zip:
        zipfile = info.get_zipfile()
        filec, dirc, totalc, totalu = zip_entry_rollup(zipfile)
        prcnt = (float(totalc) / float(totalu)) * 100

        data["zip.type"] = zipfile.__class__.__name__
        data["zip.file_count"] = filec
        data["zip.dir_count"] = dirc
        data["zip.uncompressed_size"] = totalu
        data["zip.compressed_size"] = totalc
        data["zip.compress_percent"] = prcnt

    dump(data, sys.stdout, sort_keys=True, indent=2)


def cli(options):
    if options.verbose:
        options.zip = True
        options.lines = True
        options.locals = True
        options.disassemble = True
        options.sigs = True
        options.constpool = True

    options.indent = not(options.lines or
                         options.disassemble or
                         options.sigs)

    for fn in options.jarfiles:
        with JarInfo(filename=fn) as ji:
            if options.json:
                cli_jarinfo_json(options, ji)
            else:
                cli_jarinfo(options, ji)

    return 0


def add_jarinfo_optgroup(parser):
    g = parser.add_argument_group("JAR Info Options")

    g.add_argument("--zip", action="store_true", default=False,
                   help="print zip information")

    g.add_argument("--manifest", action="store_true", default=False,
                   help="print manifest information")

    g.add_argument("--jar-classes", action="store_true", default=False,
                   help="print information about contained classes")

    g.add_argument("--jar-provides", dest="jar_provides",
                   action="store_true", default=False,
                   help="API provides information at the JAR level")

    g.add_argument("--jar-requires", dest="jar_requires",
                   action="store_true", default=False,
                   help="API requires information at the JAR level")


def create_optparser(progname):
    parser = ArgumentParser(progname)

    parser.add_argument("jarfiles", nargs="+",
                        help="JAR files to inspect")
    parser.add_argument("--json", dest="json", action="store_true",
                        help="output in JSON mode")
    add_jarinfo_optgroup(parser)
    add_classinfo_optgroup(parser)

    return parser


def main(args=sys.argv):
    parser = create_optparser(args[0])
    return cli(parser.parse_args(args[1:]))


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Utility script and module for producing a set of changes in a JAR
file. Takes the same options as the classdiff script, and in fact uses
the classdiff script's code on each non-identical member in a pair of
JAR files.

:author: Christopher O'Brien  <obriencj@gmail.com>
:licence: LGPL
"""

import sys
from argparse import ArgumentParser
from os.path import isdir

from . import unpack_class
from .change import GenericChange, SuperChange, Addition, Removal
from .change import squash, yield_sorted_by_type
from .classdiff import JavaClassChange, JavaClassReport
from .dirutils import fnmatches
from .manifest import Manifest, ManifestChange
from .manifest import SignatureManifestChange, SignatureBlockFileChange
from .manifest import file_matches_sigfile, file_matches_sigblock
from .ziputils import compare_zips, open_zip, open_zip_entry
from .ziputils import LEFT, RIGHT, DIFF, SAME


__all__ = (
    "JarChange",
    "JarTypeChange", "JarContentsChange",
    "JarManifestChange",
    "JarContentChange", "JarContentAdded", "JarContentRemoved",
    "JarSignatureFileChange", "JarSignatureFileAdded",
    "JarSignatureFileRemoved",
    "JarSignatureBlockFileChange", "JarSignatureBlockFileAdded",
    "JarSignatureBlockFileRemoved",
    "JarClassChange", "JarClassAdded", "JarClassRemoved",
    "JarReport", "JarContentsReport", "JarClassReport",
    "cli", "main",
    "cli_jars_diff",
    "add_jardiff_optgroup", "default_jardiff_options", )


class JarTypeChange(GenericChange):
    """
    exploded vs. zipped and compression level
    """

    label = "Jar type"


    def fn_data(self, c):
        # TODO: create some kind of menial zipinfo output showing type
        # (exploded/zipped) and compression level

        return isdir(c)


    def fn_pretty(self, c):
        if isdir(c):
            return "exploded JAR"
        else:
            return "zipped JAR file"


class JarContentChange(SuperChange):
    """
    a file or directory changed between JARs
    """

    label = "Jar Content"


    def __init__(self, lzip, rzip, entry, is_change=True):
        super(JarContentChange, self).__init__(lzip, rzip)
        self.entry = entry
        self.changed = is_change


    def open_left(self):
        return open_zip_entry(self.ldata, self.entry)


    def open_right(self):
        return open_zip_entry(self.rdata, self.entry)


    def collect_impl(self):
        """
        Content changes refer to more concrete children, but by default
        are empty
        """

        return tuple()


    def get_description(self):
        c = "has changed" if self.is_change() else "is unchanged"
        return "%s %s: %s" % (self.label, c, self.entry)


    def is_ignored(self, options):
        return fnmatches(self.entry, *options.ignore_jar_entry) or \
            super(JarContentChange, self).is_ignored(options)


class JarContentAdded(JarContentChange, Addition):

    label = "Jar Content Added"


    def get_description(self):
        return "%s: %s" % (self.label, self.entry)


class JarContentRemoved(JarContentChange, Removal):

    label = "Jar Content Removed"


    def get_description(self):
        return "%s: %s" % (self.label, self.entry)


class JarClassAdded(JarContentAdded):

    label = "Java Class Added"


class JarClassRemoved(JarContentRemoved):

    label = "Java Class Removed"


class JarClassChange(JarContentChange):

    label = "Java Class"


    def collect_impl(self):
        if self.is_change():
            with self.open_left() as lfd:
                linfo = unpack_class(lfd.read())

            with self.open_right() as rfd:
                rinfo = unpack_class(rfd.read())

            yield JavaClassChange(linfo, rinfo)


class JarClassReport(JarClassChange):

    report_name = "JavaClassReport"


    def __init__(self, l, r, entry, reporter):
        super(JarClassReport, self).__init__(l, r, entry)
        self.reporter = reporter


    def collect_impl(self):
        if self.is_change():
            with self.open_left() as lfd:
                linfo = unpack_class(lfd.read())

            with self.open_right() as rfd:
                rinfo = unpack_class(rfd.read())

            yield JavaClassReport(linfo, rinfo, self.reporter)


class JarManifestChange(JarContentChange):

    label = "Jar Manifest"


    def collect_impl(self):
        if self.is_change():
            with self.open_left() as lfd:
                lm = Manifest()
                lm.parse(lfd.read())

            with self.open_right() as rfd:
                rm = Manifest()
                rm.parse(rfd.read())

            yield ManifestChange(lm, rm)


class JarSignatureFileChange(JarContentChange):

    label = "Jar Signature File"

    def is_ignored(self, options):
        return options.ignore_jar_signature

    def collect_impl(self):
        if self.is_change():
            with self.open_left() as lfd:
                lm = Manifest()
                lm.parse(lfd.read())

            with self.open_right() as rfd:
                rm = Manifest()
                rm.parse(rfd.read())

            yield SignatureManifestChange(lm, rm)


class JarSignatureFileAdded(JarContentAdded):

    label = "Jar Signature File Added"

    def is_ignored(self, options):
        return options.ignore_jar_signature


class JarSignatureFileRemoved(JarContentRemoved):

    label = "Jar Signature File Removed"

    def is_ignored(self, options):
        return options.ignore_jar_signature


class JarSignatureBlockFileChange(JarContentChange):

    label = "Jar Signature Block File"

    def is_ignored(self, options):
        return options.ignore_jar_signature

    def collect_impl(self):
        if self.is_change():
            with self.open_left() as lfd, self.open_right() as rfd:
                lsig = lfd.read()
                rsig = rfd.read()
            yield SignatureBlockFileChange(lsig, rsig)


class JarSignatureBlockFileAdded(JarContentAdded):

    label = "Jar Signature Block File Added"


    def is_ignored(self, options):
        return options.ignore_jar_signature


class JarSignatureBlockFileRemoved(JarContentRemoved):

    label = "Jar Signature Block File Removed"


    def is_ignored(self, options):
        return options.ignore_jar_signature


class GenericFileChange(GenericChange):
    label = "Generic File"

    def get_description(self):
        return "[generic file change]"

    def fn_pretty(self, side_data):
        # We do not know how to represent it.
        # E.g. binary data can choke JSON encoder.
        return "[data]"


class JarGenericFileChange(JarContentChange):

    label = "Jar Generic File"

    def collect_impl(self):
        if self.is_change():
            with self.open_left() as lfd, self.open_right() as rfd:
                lsig = lfd.read()
                rsig = rfd.read()
            yield GenericFileChange(lsig, rsig)


class JarContentsChange(SuperChange):

    label = "JAR Contents"


    def __init__(self, left_fn, right_fn):
        super(JarContentsChange, self).__init__(left_fn, right_fn)
        self.lzip = None
        self.rzip = None


    @yield_sorted_by_type(JarManifestChange,
                          JarSignatureFileAdded,
                          JarSignatureFileRemoved,
                          JarSignatureFileChange,
                          JarSignatureBlockFileAdded,
                          JarSignatureBlockFileRemoved,
                          JarSignatureBlockFileChange,
                          JarGenericFileChange,
                          JarContentAdded,
                          JarContentRemoved,
                          JarContentChange,
                          JarClassAdded,
                          JarClassRemoved,
                          JarClassChange)
    def collect_impl(self):
        # these are opened for the duration of check_impl
        left = self.lzip
        right = self.rzip

        # this is our guarantee from invokation order
        assert(left is not None)
        assert(right is not None)

        for event, entry in compare_zips(left, right):
            if event == SAME:

                # TODO: should we split by file type to more specific
                # types of (un)changes? For now just emit a content
                # change with is_change set to False.

                if entry == "META-INF/MANIFEST.MF":
                    yield JarManifestChange(left, right, entry, False)

                elif file_matches_sigfile(entry):
                    yield JarSignatureFileChange(left, right, entry, False)

                elif file_matches_sigblock(entry):
                    yield JarSignatureBlockFileChange(left, right,
                                                      entry, False)

                elif fnmatches(entry, "*.class"):
                    yield JarClassChange(left, right, entry, False)

                else:
                    yield JarContentChange(left, right, entry, False)

            elif event == DIFF:
                if entry == "META-INF/MANIFEST.MF":
                    yield JarManifestChange(left, right, entry)

                elif file_matches_sigfile(entry):
                    yield JarSignatureFileChange(left, right, entry)

                elif file_matches_sigblock(entry):
                    yield JarSignatureBlockFileChange(left, right, entry)

                elif fnmatches(entry, "*.class"):
                    yield JarClassChange(left, right, entry)

                else:
                    yield JarGenericFileChange(left, right, entry)

            elif event == LEFT:
                if file_matches_sigfile(entry):
                    yield JarSignatureFileRemoved(left, right, entry)

                elif file_matches_sigblock(entry):
                    yield JarSignatureBlockFileRemoved(left, right, entry)

                elif fnmatches(entry, "*.class"):
                    yield JarClassRemoved(left, right, entry)

                else:
                    yield JarContentRemoved(left, right, entry)

            elif event == RIGHT:
                if file_matches_sigfile(entry):
                    yield JarSignatureFileAdded(left, right, entry)

                elif file_matches_sigblock(entry):
                    yield JarSignatureBlockFileAdded(left, right, entry)

                elif fnmatches(entry, "*.class"):
                    yield JarClassAdded(left, right, entry)

                else:
                    yield JarContentAdded(left, right, entry)


    def check_impl(self):
        """
        Overridden to open the left and right zipfiles and to provide all
        subchecks with an open ZipFile instance rather than having
        them all open and close the ZipFile individually. For the
        duration of the check (which calls collect_impl), the
        attributes self.lzip and self.rzip will be available and used
        as the ldata and rdata of all subchecks.
        """

        with open_zip(self.ldata) as lzip:
            with open_zip(self.rdata) as rzip:
                self.lzip = lzip
                self.rzip = rzip
                ret = super(JarContentsChange, self).check_impl()

        self.lzip = None
        self.rzip = None

        return ret


class JarChange(SuperChange):

    label = "JAR"


    change_types = (JarTypeChange,
                    JarContentsChange)


class JarContentsReport(JarContentsChange):
    """
    overridden JarContentsChange which will swap out JarClassChange
    with JarClassReport instances. The check_impl method gains the
    side effect of causing all JarClassReports gathered to write
    reports of themselves to file.
    """


    def __init__(self, left_fn, right_fn, reporter):
        super(JarContentsReport, self).__init__(left_fn, right_fn)
        self.reporter = reporter


    def collect_impl(self):
        # a filter on the collect_impl of JarContentsChange which
        # replaces JarClassChange instances with a JarClassReport
        # instance instead.

        for change in JarContentsChange.collect_impl(self):
            if isinstance(change, JarClassChange) and change.is_change():
                name = JarClassReport.report_name
                sub_r = self.reporter.subreporter(change.entry, name)
                change = JarClassReport(change.ldata, change.rdata,
                                        change.entry, sub_r)
            yield change


    def check_impl(self):
        options = self.reporter.options
        changes = list()
        c = False

        with open_zip(self.ldata) as lzip:
            with open_zip(self.rdata) as rzip:

                self.lzip = lzip
                self.rzip = rzip

                for change in self.collect_impl():
                    change.check()
                    c = c or change.is_change()

                    if isinstance(change, JarClassReport):
                        changes.append(squash(change, options=options))
                        change.clear()
                    else:
                        changes.append(change)

        self.lzip = None
        self.rzip = None

        self.changes = changes
        return c, None


class JarReport(JarChange):
    """
    This class has side-effects. Running the check method with the
    reportdir options set to True will cause the deep checks to be
    written to file in that directory
    """

    report_name = "JarReport"


    def __init__(self, l, r, reporter):
        super(JarReport, self).__init__(l, r)
        self.reporter = reporter


    def collect_impl(self):
        for c in JarChange.collect_impl(self):
            if isinstance(c, JarContentsChange):
                c = JarContentsReport(c.ldata, c.rdata, self.reporter)
            yield c


    def check(self):
        # do the actual checking
        JarChange.check(self)

        # write to file
        self.reporter.run(self)


# ---- Begin jardiff CLI ----
#


def cli_jars_diff(options, left, right):
    from .report import quick_report, Reporter
    from .report import JSONReportFormat, TextReportFormat

    reports = getattr(options, "reports", tuple())
    if reports:
        rdir = options.report_dir or "./"

        rpt = Reporter(rdir, JarReport.report_name, options)
        rpt.add_formats_by_name(reports)

        delta = JarReport(left, right, rpt)

    else:
        delta = JarChange(left, right)

    delta.check()

    if not options.silent:
        if options.json:
            quick_report(JSONReportFormat, delta, options)
        else:
            quick_report(TextReportFormat, delta, options)

    if (not delta.is_change()) or delta.is_ignored(options):
        return 0
    else:
        return 1


def cli(options):

    left, right = options.jar
    return cli_jars_diff(options, left, right)


def add_jardiff_optgroup(parser):
    """
    option group specific to the tests in jardiff
    """

    og = parser.add_argument_group("JAR Checking Options")

    og.add_argument("--ignore-jar-entry", action="append", default=[])

    og.add_argument("--ignore-jar-signature",
                    action="store_true", default=False,
                    help="Ignore JAR signing changes")

    og.add_argument("--ignore-manifest",
                    action="store_true", default=False,
                    help="Ignore changes to manifests")

    og.add_argument("--ignore-manifest-subsections",
                    action="store_true", default=False,
                    help="Ignore changes to manifest subsections")

    og.add_argument("--ignore-manifest-key",
                    action="append", default=[],
                    help="case-insensitive manifest keys to ignore")


def create_optparser(progname=None):
    """
    an OptionParser instance with the appropriate options and groups
    for the jardiff utility
    """

    from .classdiff import add_general_optgroup, add_classdiff_optgroup
    from javatools import report

    parser = ArgumentParser(prog=progname)
    parser.add_argument("jar", nargs=2,
                        help="JAR files to compare")

    add_general_optgroup(parser)
    add_jardiff_optgroup(parser)
    add_classdiff_optgroup(parser)

    report.add_general_report_optgroup(parser)
    report.add_json_report_optgroup(parser)
    report.add_html_report_optgroup(parser)

    return parser


def default_jardiff_options(updates=None):
    """
    generate an options object with the appropriate default values in
    place for API usage of jardiff features. overrides is an optional
    dictionary which will be used to update fields on the options
    object.
    """

    parser = create_optparser()
    options, _args = parser.parse_args(list())

    if updates:
        # pylint: disable=W0212
        options._update_careful(updates)

    return options


def main(args=sys.argv):
    """
    main entry point for the jardiff CLI
    """

    parser = create_optparser(args[0])
    return cli(parser.parse_args(args[1:]))


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Utility script and module for inspecting binary java class files

Let's pretend to be the javap tool shipped with many Java SDKs

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL
"""


from __future__ import print_function


import sys
import javatools.opcodes as opcodes

from argparse import ArgumentParser
from json import dump
from six.moves import range

from . import platform_from_version, unpack_classfile


__all__ = (
    "SHOW_HEADER", "SHOW_PUBLIC",
    "SHOW_PACKAGE", "SHOW_PRIVATE",
    "main", "cli", "add_classinfo_optgroup",
    "cli_class_provides", "cli_class_requires",
    "cli_json_class", "cli_print_class",
    "cli_print_classinfo", "cli_simplify_classinfo",
    "cli_simplify_field", "cli_simplify_fields",
    "cli_simplify_method", "cli_simplify_methods",
)


SHOW_HEADER = 0
SHOW_PUBLIC = 1
SHOW_PACKAGE = 3
SHOW_PRIVATE = 7


def should_show(options, member):
    """
    whether to show a member by its access flags and the show
    option. There's probably a faster and smarter way to do this, but
    eh.
    """

    show = options.show
    if show == SHOW_PUBLIC:
        return member.is_public()
    elif show == SHOW_PACKAGE:
        return member.is_public() or member.is_protected()
    elif show == SHOW_PRIVATE:
        return True


def print_field(options, field):

    if options.indent:
        print("   ", end="")

    print("%s;" % field.pretty_descriptor())

    if options.sigs:
        print("  Signature:", field.get_signature())

    if options.verbose:
        if field.get_annotations():
            print("  RuntimeVisibleAnnotations:")
            index = 0
            for anno in field.get_annotations():
                print("  %i: %s" % (index, anno.pretty_annotation()))

        cv = field.get_constantvalue()
        if cv is not None:
            t, v = field.cpool.pretty_const(cv)
            if t:
                print("  Constant value:", t, v)
        print()


def print_method(options, method):
    if options.indent:
        print("   ", end="")

    print("%s;" % method.pretty_descriptor())

    if options.sigs:
        print("  Signature:", method.get_signature())

    if method.get_annotations():
        print("  RuntimeVisibleAnnotations:")
        index = 0
        for anno in method.get_annotations():
            print("  %i: %s" % (index, anno.pretty_annotation()))

    code = method.get_code()
    if options.disassemble and code:

        print("  Code:")

        if options.verbose:
            # the arg count is the number of arguments consumed from
            # the stack when this method is called. non-static methods
            # implicitly have a "this" argument that's not in the
            # descriptor
            argsc = len(method.get_arg_type_descriptors())
            if not method.is_static():
                argsc += 1

            print("   Stack=%i, Locals=%i, Args_size=%i" %
                  (code.max_stack, code.max_locals, argsc))

        for line in code.disassemble():
            opname = opcodes.get_opname_by_code(line[1])
            args = line[2]
            if args:
                args = ", ".join(str(arg) for arg in args)
                print("   %i:\t%s\t%s" % (line[0], opname, args))
            else:
                print("   %i:\t%s" % (line[0], opname))

        exps = code.exceptions
        if exps:
            print("  Exception table:")
            print("   from\tto\ttarget\ttype")
            for e in exps:
                ctype = e.pretty_catch_type()
                print("  % 4i\t% 4i\t% 4i\t%s" %
                      (e.start_pc, e.end_pc, e.handler_pc, ctype))

    if options.verbose:
        if method.is_deprecated():
            print("  Deprecated: true")

        if method.is_synthetic():
            print("  Synthetic: true")

        if method.is_bridge():
            print("  Bridge: true")

        if method.is_varargs():
            print("  Varargs: true")

    if options.lines and code:
        lnt = method.get_code().get_linenumbertable()
        if lnt:
            print("  LineNumberTable:")
            for o, l in lnt:
                print("   line %i: %i" % (l, o))

    if options.locals and code:
        if method.cpool:
            cval = method.cpool.deref_const
        else:
            cval = str

        lvt = method.get_code().get_localvariabletable()
        lvtt = method.get_code().get_localvariabletypetable()

        if lvt:
            print("  LocalVariableTable:")
            print("   Start  Length  Slot\tName\tDescriptor")
            for o, l, n, d, i in lvt:
                line = (str(o), str(l), str(i), cval(n), cval(d))
                print("   %s" % "\t".join(line))

        if lvtt:
            print("  LocalVariableTypeTable:")
            print("   Start  Length  Slot\tName\tSignature")
            for o, l, n, s, i in lvtt:
                line = (str(o), str(l), str(i), cval(n), cval(s))
                print("   %s" % "\t".join(line))

    if options.verbose:
        exps = method.pretty_exceptions()
        if exps:
            print("  Exceptions:")
            for e in exps:
                print("   throws", e)

        print()


def cli_class_provides(options, info):
    print("class %s provides:" % info.pretty_this())

    for provided in sorted(info.get_provides(options.api_ignore)):
        print(" ", provided)
    print()


def cli_class_requires(options, info):
    print("class %s requires:" % info.pretty_this())

    for required in sorted(info.get_requires(options.api_ignore)):
        print(" ", required)
    print()


def cli_print_classinfo(options, info):
    if options.class_provides or options.class_requires:
        if options.class_provides:
            cli_class_provides(options, info)
        if options.class_requires:
            cli_class_requires(options, info)
        return

    sourcefile = info.get_sourcefile()
    if sourcefile:
        print("Compiled from \"%s\"" % sourcefile)

    print(info.pretty_descriptor(), end="")

    if options.verbose or options.show == SHOW_HEADER:
        print()
        if info.get_sourcefile():
            print("  SourceFile: \"%s\"" % info.get_sourcefile())
        if info.get_signature():
            print("  Signature:", info.get_signature())

        if info.get_annotations():
            print("  RuntimeVisibleAnnotations:")
            index = 0
            for anno in info.get_annotations():
                print("  %i: %s" % (index, anno.pretty_annotation()))

        if info.get_enclosingmethod():
            print("  EnclosingMethod:", info.get_enclosingmethod())
        major, minor = info.get_version()
        print("  minor version:", major)
        print("  major version:", minor)
        platform = platform_from_version(*info.version) or "unknown"
        print("  Platform:", platform)

    if options.constpool:
        print("  Constants pool:")

        # we don't use the info.pretty_constants() generator here
        # because we actually want numbers for the entries, and that
        # generator skips them.
        cpool = info.cpool

        for i in range(1, len(cpool.consts)):
            t, v = cpool.pretty_const(i)
            if t:
                # skipping the None consts, which would be the entries
                # comprising the second half of a long or double value
                print("const #%i = %s\t%s;" % (i, t, v))
        print()

    if options.show == SHOW_HEADER:
        return

    print("{")

    for field in info.fields:
        if should_show(options, field):
            print_field(options, field)

    for method in info.methods:
        if should_show(options, method):
            print_method(options, method)

    print("}\n")

    return 0


def cli_print_class(options, classfile):
    info = unpack_classfile(classfile)
    return cli_print_classinfo(options, info)


def cli_simplify_field(field, data=None):
    if data is None:
        data = dict()

    data["name"] = field.get_name()
    data["type"] = field.pretty_type()
    data["access_flags"] = tuple(field.pretty_access_flags())

    ifonly(data, "signature", field.get_signature())
    ifonly(data, "deprecated", field.is_deprecated())

    cv = field.get_constantvalue()
    if cv is not None:
        t, v = field.cpool.pretty_const(cv)
        if t:
            data["constant_value"] = (t, v)

    return data


def cli_simplify_method(method, data=None):
    if data is None:
        data = dict()

    data["name"] = method.get_name()
    data["type"] = method.pretty_type()
    data["access_flags"] = tuple(method.pretty_access_flags())
    data["arg_types"] = tuple(method.pretty_arg_types())

    ifonly(data, "signature", method.get_signature())
    ifonly(data, "deprecated", method.is_deprecated())

    return data


def cli_simplify_fields(options, info):
    fields = list()
    for field in info.fields:
        if should_show(options, field):
            fields.append(cli_simplify_field(options, field))
    return fields


def cli_simplify_methods(options, info):
    methods = list()
    for method in info.methods:
        if should_show(options, method):
            methods.append(cli_simplify_method(options, method))
    return methods


def cli_simplify_classinfo(options, info, data=None):
    if data is None:
        data = dict()

    if options.class_provides:
        data["class_provides"] = info.get_provides(options.api_ignore)
    if options.class_requires:
        data["class_requires"] = info.get_requires(options.api_ignore)

    data["name"] = info.pretty_this()
    data["extends"] = info.pretty_super()
    data["implements"] = tuple(info.pretty_interfaces())
    data["source_file"] = info.get_sourcefile()

    ifonly(data, "signature", info.get_signature())
    ifonly(data, "enclosing_method", info.get_enclosingmethod())

    data["version"] = info.get_version()
    data["platform"] = platform_from_version(*info.version)

    if options.constpool:
        data["constants_pool"] = tuple(info.cpool.pretty_constants())

    data["fields"] = cli_simplify_fields(options, info)
    data["methods"] = cli_simplify_methods(options, info)

    return data


def ifonly(data, key, val):

    """ utility function to set data[key] to val, but only if val has
    a truthy value """

    if val:
        data[key] = val


def cli_json_class(options, classfile):
    info = unpack_classfile(classfile)
    data = cli_simplify_classinfo(options, info)
    dump(data, sys.stdout, sort_keys=True, indent=2)


def cli(options):
    if options.verbose:
        # verbose also sets all of the following options
        options.lines = True
        options.locals = True
        options.disassemble = True
        options.sigs = True
        options.constpool = True

    # just a tiny hack to mimic some indenting sun's javap will do if
    # the output is terse
    options.indent = not(options.lines or
                         options.disassemble or
                         options.sigs)

    style = cli_print_class
    if options.json:
        style = cli_json_class

    for f in options.classfile:
        style(options, f)

    return 0


def add_classinfo_optgroup(parser):
    g = parser.add_argument_group("Class Info Options")

    g.add_argument("--class-provides", dest="class_provides",
                   action="store_true", default=False,
                   help="API provides information at the class level")

    g.add_argument("--class-requires", dest="class_requires",
                   action="store_true", default=False,
                   help="API requires information at the class level")

    g.add_argument("--api-ignore", dest="api_ignore",
                   action="append", default=list(),
                   help="globs of packages to not print in provides"
                   " or requires modes")

    g.add_argument("--header", dest="show",
                   action="store_const", default=SHOW_PUBLIC,
                   const=SHOW_HEADER,
                   help="show just the class header, no members")

    g.add_argument("--public", dest="show",
                   action="store_const", const=SHOW_PUBLIC,
                   help="show only public members")

    g.add_argument("--package", dest="show",
                   action="store_const", const=SHOW_PACKAGE,
                   help="show public and protected members")

    g.add_argument("--private", dest="show",
                   action="store_const", const=SHOW_PRIVATE,
                   help="show all members")

    g.add_argument("-l", dest="lines", action="store_true",
                   help="show the line number table")

    g.add_argument("-o", dest="locals", action="store_true",
                   help="show the local variable tables")

    g.add_argument("-c", dest="disassemble", action="store_true",
                   help="disassemble method code")

    g.add_argument("-s", dest="sigs", action="store_true",
                   help="show internal type signatures")

    g.add_argument("-p", dest="constpool", action="store_true",
                   help="show the constants pool")

    g.add_argument("--verbose", dest="verbose", action="store_true",
                   help="sets -locsp options and shows stack bounds")


def create_optparser(progname):
    parser = ArgumentParser(prog=progname)
    parser.add_argument("classfile", nargs="+",
                        help="Java class file(s) to inspect")
    parser.add_argument("--json", action="store_true", default=False,
                        help="output JSON")

    add_classinfo_optgroup(parser)

    return parser


def main(args=sys.argv):
    parser = create_optparser(args[0])
    return cli(parser.parse_args(args[1:]))


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.



"""
Collects available Cheetah templates for use with the HTML report

author: Christopher O'Brien <obriencj@gmail.com>
license: LGPL
"""


from pkgutil import iter_modules


def _iter_templates():
    """
    uses reflection to yield the Cheetah templates under this module
    """

    # pylint: disable=W0406
    # needed for introspection
    import javatools.cheetah
    from Cheetah.Template import Template

    for _, name, _ in iter_modules(__path__):
        if name == "setuptools":
            continue

        __import__("javatools.cheetah." + name)
        found = getattr(getattr(javatools.cheetah, name), name)
        if issubclass(found, Template):
            yield found


def get_templates():
    """
    The Cheetah Template classes contained within this module
    """

    return tuple(_iter_templates())


def xml_entity_escape(data):
    """
    replace special characters with their XML entity versions
    """

    data = data.replace("&", "&amp;")
    data = data.replace(">", "&gt;")
    data = data.replace("<", "&lt;")
    return data


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
A module to hold all the Java opcodes. Data taken from publicly
available sources (see following for more information)

References
----------
* http://java.sun.com/docs/books/jvms/second_edition/html/VMSpecTOC.doc.html

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL v.3
"""


from functools import partial
from six.moves import range

from .pack import compile_struct


__all__ = (
    "get_opcode_by_name", "get_opname_by_code",
    "get_arg_format", "has_const_arg",
    "disassemble",
    "OP_aaload", "OP_aastore", "OP_aconst_null", "OP_aload", "OP_aload_0",
    "OP_aload_1", "OP_aload_2", "OP_aload_3", "OP_anewarray", "OP_areturn",
    "OP_arraylength", "OP_astore", "OP_astore_0", "OP_astore_1",
    "OP_astore_2", "OP_astore_3", "OP_athrow", "OP_baload", "OP_bastore",
    "OP_bipush", "OP_caload", "OP_castore", "OP_checkcast", "OP_d2f",
    "OP_d2i", "OP_d2l", "OP_dadd", "OP_daload", "OP_dastore", "OP_dcmpg",
    "OP_dcmpl", "OP_dconst_0", "OP_dconst_1", "OP_ddiv", "OP_dload",
    "OP_dload_0", "OP_dload_1", "OP_dload_2", "OP_dload_3", "OP_dmul",
    "OP_dneg", "OP_drem", "OP_dreturn", "OP_dstore", "OP_dstore_0",
    "OP_dstore_1", "OP_dstore_2", "OP_dstore_3", "OP_dsub", "OP_dup",
    "OP_dup2", "OP_dup2_x1", "OP_dup2_x2", "OP_dup_x1", "OP_dup_x2",
    "OP_f2d", "OP_f2i", "OP_f2l", "OP_fadd", "OP_faload", "OP_fastore",
    "OP_fcmpg", "OP_fcmpl", "OP_fconst_0", "OP_fconst_1", "OP_fconst_2",
    "OP_fdiv", "OP_fload", "OP_fload_0", "OP_fload_1", "OP_fload_2",
    "OP_fload_3", "OP_fmul", "OP_fneg", "OP_frem", "OP_freturn",
    "OP_fstore", "OP_fstore_0", "OP_fstore_1", "OP_fstore_2",
    "OP_fstore_3", "OP_fsub", "OP_getfield", "OP_getstatic", "OP_goto",
    "OP_goto_w", "OP_i2b", "OP_i2c", "OP_i2d", "OP_i2f", "OP_i2l", "OP_i2s",
    "OP_iadd", "OP_iaload", "OP_iand", "OP_iastore", "OP_iconst_0",
    "OP_iconst_1", "OP_iconst_2", "OP_iconst_3", "OP_iconst_4",
    "OP_iconst_5", "OP_iconst_m1", "OP_idiv", "OP_if_acmpeq",
    "OP_if_acmpne", "OP_if_icmpeq", "OP_if_icmpge", "OP_if_icmpgt",
    "OP_if_icmple", "OP_if_icmplt", "OP_if_icmpne", "OP_ifeq", "OP_ifge",
    "OP_ifgt", "OP_ifle", "OP_iflt", "OP_ifne", "OP_ifnonnull",
    "OP_ifnull", "OP_iinc", "OP_iload", "OP_iload_0", "OP_iload_1",
    "OP_iload_2", "OP_iload_3", "OP_imul", "OP_ineg", "OP_instanceof",
    "OP_invokedynamic", "OP_invokeinterface", "OP_invokespecial",
    "OP_invokestatic", "OP_invokevirtual", "OP_ior", "OP_irem",
    "OP_ireturn", "OP_ishl", "OP_ishr", "OP_istore", "OP_istore_0",
    "OP_istore_1", "OP_istore_2", "OP_istore_3", "OP_isub", "OP_iushr",
    "OP_ixor", "OP_jsr", "OP_jsr_w", "OP_l2d", "OP_l2f", "OP_l2i",
    "OP_ladd", "OP_laload", "OP_land", "OP_lastore", "OP_lcmp",
    "OP_lconst_0", "OP_lconst_1", "OP_ldc", "OP_ldc2_w", "OP_ldc_w",
    "OP_ldiv", "OP_lload", "OP_lload_0", "OP_lload_1", "OP_lload_2",
    "OP_lload_3", "OP_lmul", "OP_lneg", "OP_lookupswitch", "OP_lor",
    "OP_lrem", "OP_lreturn", "OP_lshl", "OP_lshr", "OP_lstore",
    "OP_lstore_0", "OP_lstore_1", "OP_lstore_2", "OP_lstore_3", "OP_lsub",
    "OP_lushr", "OP_lxor", "OP_monitorentry", "OP_monitorexit",
    "OP_multianewarray", "OP_new", "OP_newarray", "OP_nop", "OP_pop",
    "OP_pop2", "OP_putfield", "OP_putstatic", "OP_ret", "OP_return",
    "OP_saload", "OP_sastore", "OP_sipush", "OP_swap", "OP_tableswitch",
    "OP_wide",
)


# the op table itself
__OPTABLE = {}

# mnemonics for the op tuples
_OPINDEX_NAME = 0
_OPINDEX_VAL = 1
_OPINDEX_FMT = 2
_OPINDEX_CONSUME = 3
_OPINDEX_PRODUCE = 4
_OPINDEX_CONST = 5


# commonly re-occurring struct formats
# pylint: disable=C0103
_struct_i = compile_struct(">i")
_struct_ii = compile_struct(">ii")
_struct_iii = compile_struct(">iii")
_struct_BH = compile_struct(">BH")
_struct_BHh = compile_struct(">BHh")


def __op(name, val, fmt=None, const=False, consume=0, produce=0):
    """
    provides sensible defaults for a code, and registers it with the
    __OPTABLE for lookup.
    """

    name = name.lower()

    # fmt can either be a str representing the struct to unpack, or a
    # callable to do more complex unpacking. If it's a str, create a
    # callable for it.
    if isinstance(fmt, str):
        fmt = partial(_unpack, compile_struct(fmt))

    operand = (name, val, fmt, consume, produce, const)

    assert(name not in __OPTABLE)
    assert(val not in __OPTABLE)

    __OPTABLE[name] = operand
    __OPTABLE[val] = operand

    return val


def get_opcode_by_name(name):
    """
    get the integer opcode by its name
    """

    return __OPTABLE[name.lower()][_OPINDEX_VAL]


def get_opname_by_code(code):
    """
    get the name of an opcode
    """

    return __OPTABLE[code][_OPINDEX_NAME]


def get_arg_format(code):
    """
    get the format of arguments to this opcode
    """

    return __OPTABLE[code][_OPINDEX_FMT]


def has_const_arg(code):
    """
    which arg is a const for this opcode
    """

    return __OPTABLE[code][_OPINDEX_CONST]


def _unpack(struct, bc, offset=0):
    """
    returns the unpacked data tuple, and the next offset past the
    unpacked data
    """

    return struct.unpack_from(bc, offset), offset + struct.size


def _unpack_lookupswitch(bc, offset):
    """
    function for unpacking the lookupswitch op arguments
    """

    jump = (offset % 4)
    if jump:
        offset += (4 - jump)

    (default, npairs), offset = _unpack(_struct_ii, bc, offset)

    switches = list()
    for _index in range(npairs):
        pair, offset = _unpack(_struct_ii, bc, offset)
        switches.append(pair)

    return (default, switches), offset


def _unpack_tableswitch(bc, offset):
    """
    function for unpacking the tableswitch op arguments
    """

    jump = (offset % 4)
    if jump:
        offset += (4 - jump)

    (default, low, high), offset = _unpack(_struct_iii, bc, offset)

    joffs = list()
    for _index in range((high - low) + 1):
        j, offset = _unpack(_struct_i, bc, offset)
        joffs.append(j)

    return (default, low, high, joffs), offset


def _unpack_wide(bc, offset):
    """
    unpacker for wide ops
    """

    code = ord(bc[offset])

    if code == OP_iinc:
        return _unpack(_struct_BHh, bc, offset)

    elif code in (OP_iload, OP_fload, OP_aload, OP_lload, OP_dload,
                  OP_istore, OP_fstore, OP_astore, OP_lstore,
                  OP_dstore, OP_ret):

        return _unpack(_struct_BH, bc, offset)

    else:
        # no other opcodes are valid, so shouldn't have fallen through
        # to here.
        assert False


def disassemble(bytecode):
    """
    Generator. Disassembles Java bytecode into a sequence of (offset,
    code, args) tuples
    :type bytecode: bytes
    """

    offset = 0
    end = len(bytecode)

    while offset < end:
        orig_offset = offset

        code = bytecode[offset]
        if not isinstance(code, int):   # Py3
            code = ord(code)

        offset += 1

        args = tuple()
        fmt = get_arg_format(code)
        if fmt:
            args, offset = fmt(bytecode, offset)

        yield (orig_offset, code, args)


# And now, the OP codes themselves

# The individual OP_* constants just have the numerical value. The
# rest is just information to get stored in the __OPTABLE

# pylint: disable=C0103

OP_aaload = __op('aaload', 0x32)
OP_aastore = __op('aastore', 0x53)
OP_aconst_null = __op('aconst_null', 0x1)
OP_aload = __op('aload', 0x19, fmt='>B')
OP_aload_0 = __op('aload_0', 0x2a)
OP_aload_1 = __op('aload_1', 0x2b)
OP_aload_2 = __op('aload_2', 0x2c)
OP_aload_3 = __op('aload_3', 0x2d)
OP_anewarray = __op('anewarray', 0xbd, fmt='>H', const=True)
OP_areturn = __op('areturn', 0xb0)
OP_arraylength = __op('arraylength', 0xbe)
OP_astore = __op('astore', 0x3a, fmt='>B')
OP_astore_0 = __op('astore_0', 0x4b)
OP_astore_1 = __op('astore_1', 0x4c)
OP_astore_2 = __op('astore_2', 0x4d)
OP_astore_3 = __op('astore_3', 0x4e)
OP_athrow = __op('athrow', 0xbf)

OP_baload = __op('baload', 0x33)
OP_bastore = __op('bastore', 0x54)
OP_bipush = __op('bipush', 0x10, fmt='>B')

OP_caload = __op('caload', 0x34)
OP_castore = __op('castore', 0x55)
OP_checkcast = __op('checkcast', 0xc0, fmt='>H', const=True)

OP_d2f = __op('d2f', 0x90)
OP_d2i = __op('d2i', 0x8e)
OP_d2l = __op('d2l', 0x8f)
OP_dadd = __op('dadd', 0x63)
OP_daload = __op('daload', 0x31)
OP_dastore = __op('dastore', 0x52)
OP_dcmpg = __op('dcmpg', 0x98)
OP_dcmpl = __op('dcmpl', 0x97)
OP_dconst_0 = __op('dconst_0', 0xe)
OP_dconst_1 = __op('dconst_1', 0xf)
OP_ddiv = __op('ddiv', 0x6f)
OP_dload = __op('dload', 0x18, fmt='>B')
OP_dload_0 = __op('dload_0', 0x26)
OP_dload_1 = __op('dload_1', 0x27)
OP_dload_2 = __op('dload_2', 0x28)
OP_dload_3 = __op('dload_3', 0x29)
OP_dmul = __op('dmul', 0x6b)
OP_dneg = __op('dneg', 0x77)
OP_drem = __op('drem', 0x73)
OP_dreturn = __op('dreturn', 0xaf)
OP_dstore = __op('dstore', 0x39, fmt='>B')
OP_dstore_0 = __op('dstore_0', 0x47)
OP_dstore_1 = __op('dstore_1', 0x48)
OP_dstore_2 = __op('dstore_2', 0x49)
OP_dstore_3 = __op('dstore_3', 0x4a)
OP_dsub = __op('dsub', 0x67)
OP_dup = __op('dup', 0x59)
OP_dup_x1 = __op('dup_x1', 0x5a)
OP_dup_x2 = __op('dup_x2', 0x5b)
OP_dup2 = __op('dup2', 0x5c)
OP_dup2_x1 = __op('dup2_x1', 0x5d)
OP_dup2_x2 = __op('dup2_x2', 0x5e)

OP_f2d = __op('f2d', 0x8d)
OP_f2i = __op('f2i', 0x8b)
OP_f2l = __op('f2l', 0x8c)
OP_fadd = __op('fadd', 0x62)
OP_faload = __op('faload', 0x30)
OP_fastore = __op('fastore', 0x51)
OP_fcmpg = __op('fcmpg', 0x96)
OP_fcmpl = __op('fcmpl', 0x95)
OP_fconst_0 = __op('fconst_0', 0xb)
OP_fconst_1 = __op('fconst_1', 0xc)
OP_fconst_2 = __op('fconst_2', 0xd)
OP_fdiv = __op('fdiv', 0x6e)
OP_fload = __op('fload', 0x17, fmt='>B')
OP_fload_0 = __op('fload_0', 0x22)
OP_fload_1 = __op('fload_1', 0x23)
OP_fload_2 = __op('fload_2', 0x24)
OP_fload_3 = __op('fload_3', 0x25)
OP_fmul = __op('fmul', 0x6a)
OP_fneg = __op('fneg', 0x76)
OP_frem = __op('frem', 0x72)
OP_freturn = __op('freturn', 0xae)
OP_fstore = __op('fstore', 0x38, fmt='>B')
OP_fstore_0 = __op('fstore_0', 0x43)
OP_fstore_1 = __op('fstore_1', 0x44)
OP_fstore_2 = __op('fstore_2', 0x45)
OP_fstore_3 = __op('fstore_3', 0x46)
OP_fsub = __op('fsub', 0x66)

OP_getfield = __op('getfield', 0xb4, fmt='>H', const=True)
OP_getstatic = __op('getstatic', 0xb2, fmt='>H', const=True)
OP_goto = __op('goto', 0xa7, fmt='>h')
OP_goto_w = __op('goto_w', 0xc8, fmt='>i')

OP_i2b = __op('i2b', 0x91)
OP_i2c = __op('i2c', 0x92)
OP_i2d = __op('i2d', 0x87)
OP_i2f = __op('i2f', 0x86)
OP_i2l = __op('i2l', 0x85)
OP_i2s = __op('i2s', 0x93)
OP_iadd = __op('iadd', 0x60)
OP_iaload = __op('iaload', 0x2e)
OP_iand = __op('iand', 0x7e)
OP_iastore = __op('iastore', 0x4f)
OP_iconst_m1 = __op('iconst_m1', 0x2)
OP_iconst_0 = __op('iconst_0', 0x3)
OP_iconst_1 = __op('iconst_1', 0x4)
OP_iconst_2 = __op('iconst_2', 0x5)
OP_iconst_3 = __op('iconst_3', 0x6)
OP_iconst_4 = __op('iconst_4', 0x7)
OP_iconst_5 = __op('iconst_5', 0x8)
OP_idiv = __op('idiv', 0x6c)
OP_if_acmpeq = __op('if_acmpeq', 0xa5, fmt='>h')
OP_if_acmpne = __op('if_acmpne', 0xa6, fmt='>h')
OP_if_icmpeq = __op('if_icmpeq', 0x9f, fmt='>h')
OP_if_icmpne = __op('if_icmpne', 0xa0, fmt='>h')
OP_if_icmplt = __op('if_icmplt', 0xa1, fmt='>h')
OP_if_icmpge = __op('if_icmpge', 0xa2, fmt='>h')
OP_if_icmpgt = __op('if_icmpgt', 0xa3, fmt='>h')
OP_if_icmple = __op('if_icmple', 0xa4, fmt='>h')
OP_ifeq = __op('ifeq', 0x99, fmt='>h')
OP_ifne = __op('ifne', 0x9a, fmt='>h')
OP_iflt = __op('iflt', 0x9b, fmt='>h')
OP_ifge = __op('ifge', 0x9c, fmt='>h')
OP_ifgt = __op('ifgt', 0x9d, fmt='>h')
OP_ifle = __op('ifle', 0x9e, fmt='>h')
OP_ifnonnull = __op('ifnonnull', 0xc7, fmt='>h')
OP_ifnull = __op('ifnull', 0xc6, fmt='>h')
OP_iinc = __op('iinc', 0x84, fmt='>Bb')
OP_iload = __op('iload', 0x15, fmt='>B')
OP_iload_0 = __op('iload_0', 0x1a)
OP_iload_1 = __op('iload_1', 0x1b)
OP_iload_2 = __op('iload_2', 0x1c)
OP_iload_3 = __op('iload_3', 0x1d)
OP_imul = __op('imul', 0x68)
OP_ineg = __op('ineg', 0x74)
OP_instanceof = __op('instanceof', 0xc1, fmt='>H', const=True)
OP_invokedynamic = __op('invokedynamic', 0xba, fmt='>HBB', const=True)
OP_invokeinterface = __op('invokeinterface', 0xb9, fmt='>HBB', const=True)
OP_invokespecial = __op('invokespecial', 0xb7, fmt='>H', const=True)
OP_invokestatic = __op('invokestatic', 0xb8, fmt='>H', const=True)
OP_invokevirtual = __op('invokevirtual', 0xb6, fmt='>H', const=True)
OP_ior = __op('ior', 0x80)
OP_irem = __op('irem', 0x70)
OP_ireturn = __op('ireturn', 0xac)
OP_ishl = __op('ishl', 0x78)
OP_ishr = __op('ishr', 0x7a)
OP_istore = __op('istore', 0x36, fmt='>B')
OP_istore_0 = __op('istore_0', 0x3b)
OP_istore_1 = __op('istore_1', 0x3c)
OP_istore_2 = __op('istore_2', 0x3d)
OP_istore_3 = __op('istore_3', 0x3e)
OP_isub = __op('isub', 0x64)
OP_iushr = __op('iushr', 0x7c)
OP_ixor = __op('ixor', 0x82)

OP_jsr = __op('jsr', 0xa8, fmt='>h')
OP_jsr_w = __op('jsr_w', 0xc9, fmt='>i')

OP_l2d = __op('l2d', 0x8a)
OP_l2f = __op('l2f', 0x89)
OP_l2i = __op('l2i', 0x88)
OP_ladd = __op('ladd', 0x61)
OP_laload = __op('laload', 0x2f)
OP_land = __op('land', 0x7f)
OP_lastore = __op('lastore', 0x50)
OP_lcmp = __op('lcmp', 0x94)
OP_lconst_0 = __op('lconst_0', 0x9)
OP_lconst_1 = __op('lconst_1', 0xa)
OP_ldc = __op('ldc', 0x12, fmt='>B', const=True)
OP_ldc_w = __op('ldc_w', 0x13, fmt='>H', const=True)
OP_ldc2_w = __op('ldc2_w', 0x14, fmt='>H', const=True)
OP_ldiv = __op('ldiv', 0x6d)
OP_lload = __op('lload', 0x16, fmt='>B')
OP_lload_0 = __op('lload_0', 0x1e)
OP_lload_1 = __op('lload_1', 0x1f)
OP_lload_2 = __op('lload_2', 0x20)
OP_lload_3 = __op('lload_3', 0x21)
OP_lmul = __op('lmul', 0x69)
OP_lneg = __op('lneg', 0x75)
OP_lookupswitch = __op('lookupswitch', 0xab, fmt=_unpack_lookupswitch)
OP_lor = __op('lor', 0x81)
OP_lrem = __op('lrem', 0x71)
OP_lreturn = __op('lreturn', 0xad)
OP_lshl = __op('lshl', 0x79)
OP_lshr = __op('lshr', 0x7b)
OP_lstore = __op('lstore', 0x37, fmt='>B')
OP_lstore_0 = __op('lstore_0', 0x3f)
OP_lstore_1 = __op('lstore_1', 0x40)
OP_lstore_2 = __op('lstore_2', 0x41)
OP_lstore_3 = __op('lstore_3', 0x42)
OP_lsub = __op('lsub', 0x65)
OP_lushr = __op('lushr', 0x7d)
OP_lxor = __op('lxor', 0x83)

OP_monitorentry = __op('monitorentry', 0xc2)
OP_monitorexit = __op('monitorexit', 0xc3)
OP_multianewarray = __op('multianewarray', 0xc5, fmt='>HB')

OP_new = __op('new', 0xbb, fmt='>H', const=True)
OP_newarray = __op('newarray', 0xbc, fmt='>B')
OP_nop = __op('nop', 0x0)

OP_pop = __op('pop', 0x57)
OP_pop2 = __op('pop2', 0x58)
OP_putfield = __op('putfield', 0xb5, fmt='>H', const=True)
OP_putstatic = __op('putstatic', 0xb3, fmt='>H', const=True)

OP_ret = __op('ret', 0xa9, fmt='>B')
OP_return = __op('return', 0xb1)

OP_saload = __op('saload', 0x35)
OP_sastore = __op('sastore', 0x56)
OP_sipush = __op('sipush', 0x11, fmt='>h')
OP_swap = __op('swap', 0x5f)

OP_tableswitch = __op('tableswitch', 0xaa, fmt=_unpack_tableswitch)

OP_wide = __op('wide', 0xc4, fmt=_unpack_wide)


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Utility module for discovering the differences between two directory
trees

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL
"""


from filecmp import dircmp
from fnmatch import fnmatch
from os import makedirs, walk
from os.path import exists, isdir, join, relpath
from shutil import copy


LEFT = "left only"
RIGHT = "right only"
DIFF = "changed"
SAME = "same"
BOTH = SAME  # meh, synonyms


def fnmatches(entry, *pattern_list):
    """
    returns true if entry matches any of the glob patterns, false
    otherwise
    """

    for pattern in pattern_list:
        if pattern and fnmatch(entry, pattern):
            return True
    return False


def makedirsp(dirname):
    """
    create dirname if it doesn't exist
    """

    if dirname and not exists(dirname):
        makedirs(dirname)


def copydir(orig, dest):
    """
    copies directory orig to dest. Returns a list of tuples of
    relative filenames which were copied from orig to dest
    """

    copied = list()

    makedirsp(dest)

    for root, dirs, files in walk(orig):
        for d in dirs:
            # ensure directories exist
            makedirsp(join(dest, d))

        for f in files:
            root_f = join(root, f)
            dest_f = join(dest, relpath(root_f, orig))
            copy(root_f, dest_f)
            copied.append((root_f, dest_f))

    return copied


def compare(left, right):
    """
    generator emiting pairs indicating the contents of the left and
    right directories. The pairs are in the form of (difference,
    filename) where difference is one of the LEFT, RIGHT, DIFF, or
    BOTH constants. This generator recursively walks both trees.
    """

    dc = dircmp(left, right, ignore=[])
    return _gen_from_dircmp(dc, left, right)


def _gen_from_dircmp(dc, lpath, rpath):
    """
    do the work of comparing the dircmp
    """

    left_only = dc.left_only
    left_only.sort()

    for f in left_only:
        fp = join(dc.left, f)
        if isdir(fp):
            for r, _ds, fs in walk(fp):
                r = relpath(r, lpath)
                for f in fs:
                    yield(LEFT, join(r, f))
        else:
            yield (LEFT, relpath(fp, lpath))

    right_only = dc.right_only
    right_only.sort()

    for f in right_only:
        fp = join(dc.right, f)
        if isdir(fp):
            for r, _ds, fs in walk(fp):
                r = relpath(r, rpath)
                for f in fs:
                    yield(RIGHT, join(r, f))
        else:
            yield (RIGHT, relpath(fp, rpath))

    diff_files = dc.diff_files
    diff_files.sort()

    for f in diff_files:
        yield (DIFF, join(relpath(dc.right, rpath), f))

    same_files = dc.same_files
    same_files.sort()

    for f in same_files:
        yield (BOTH, join(relpath(dc.left, lpath), f))

    subdirs = dc.subdirs.values()
    subdirs = sorted(subdirs)
    for sub in subdirs:
        for event in _gen_from_dircmp(sub, lpath, rpath):
            yield event


def collect_compare(left, right):
    """
    returns a tuple of four lists describing the file paths that have
    been (in order) added, removed, altered, or left the same
    """

    return collect_compare_into(left, right, [], [], [], [])


def collect_compare_into(left, right, added, removed, altered, same):
    """
    collect the results of compare into the given lists (or None if
    you do not wish to collect results of that type. Returns a tuple
    of (added, removed, altered, same)
    """

    for event, filename in compare(left, right):
        if event == LEFT:
            group = removed

        elif event == RIGHT:
            group = added

        elif event == DIFF:
            group = altered

        elif event == BOTH:
            group = same

        else:
            assert False

        if group is not None:
            group.append(filename)

    return added, removed, altered, same


class ClosingContext(object):
    # pylint: disable=R0903
    # too few public methods (none)

    """
    A simple context manager which is created with an object instance,
    and will return that instance from __enter__ and call the close
    method on the instance in __exit__
    """


    def __init__(self, managed):
        self.managed = managed


    def __enter__(self):
        return self.managed


    def __exit__(self, exc_type, _exc_val, _exc_tb):
        managed = self.managed
        self.managed = None

        if managed is not None and hasattr(managed, "close"):
            managed.close()

        return exc_type is None


def closing(managed):
    """
    If the managed object already provides its own context management
    via the __enter__ and __exit__ methods, it is returned
    unchanged. However, if the instance does not, a ClosingContext
    will be created to wrap it. When the ClosingContext exits, it will
    call managed.close()
    """

    if managed is None or hasattr(managed, "__enter__"):
        return managed
    else:
        return ClosingContext(managed)


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Utilities for discovering entry deltas in a pair of zip files.

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL
"""


from os import walk
from os.path import getsize, isdir, isfile, islink, join, relpath
from six import BytesIO
from six.moves import zip_longest
from zipfile import is_zipfile, ZipFile, ZipInfo, _EndRecData
from zlib import crc32

from .dirutils import LEFT, RIGHT, DIFF, SAME, closing


__all__ = (
    "compare", "compare_zips",
    "open_zip", "open_zip_entry",
    "zip_file", "zip_entry_rollup",
    "LEFT", "RIGHT", "DIFF", "SAME", )


try:
    buffer
except NameError:
    buffer = memoryview


_CHUNKSIZE = 2 ** 14


def compare(left, right):
    """
    yields EVENT,ENTRY pairs describing the differences between left
    and right, which are filenames for a pair of zip files
    """

    with open_zip(left) as lz, open_zip(right) as rz:
        return compare_zips(lz, rz)


def compare_zips(left, right):
    """
    yields EVENT,ENTRY pairs describing the differences between left
    and right ZipFile instances
    """

    ll = set(left.namelist())
    rl = set(right.namelist())

    for f in ll:
        if f in rl:
            rl.remove(f)

            if f[-1] == '/':
                # it's a directory entry
                pass

            elif _different(left, right, f):
                yield DIFF, f

            else:
                yield SAME, f

        else:
            yield LEFT, f

    for f in rl:
        yield RIGHT, f


def _different(left, right, f):
    """
    true if entry f is different between left and right ZipFile
    instances
    """

    li = left.getinfo(f)
    ri = right.getinfo(f)

    if (li.file_size == ri.file_size) and (li.CRC == ri.CRC):
        # ok, they seem passibly similar, let's deep check them.
        return _deep_different(left, right, f)

    else:
        # yup, they're different
        return True


def _deep_different(left, right, entry):
    """
    checks that entry is identical between ZipFile instances left and
    right
    """

    left = chunk_zip_entry(left, entry)
    right = chunk_zip_entry(right, entry)

    for ldata, rdata in zip_longest(left, right):
        if ldata != rdata:
            return True
    return False


def collect_compare(left, right):
    """
    collects the differences between left and right, which are
    filenames for valid zip files, into a tuple of lists: added,
    removed, altered, same
    """

    return collect_compare_into(left, right, [], [], [], [])


def collect_compare_into(left, right, added, removed, altered, same):
    """
    collects the differences between left and right, which are
    filenames for valid zip files, into the lists added, removed,
    altered, and same.  Returns a tuple of added, removed, altered,
    same
    """

    with open_zip(left) as lz, open_zip(right) as rz:
        return collect_compare_zips_into(lz, rz,
                                         added, removed,
                                         altered, same)


def collect_compare_zips(left, right):
    """
    collects the differences between left and right ZipFile instances
    into a tuple of lists: added, removed, altered, same
    """

    return collect_compare_zips_into(left, right, [], [], [], [])


def collect_compare_zips_into(left, right, added, removed, altered, same):
    """
    collects the differences between left and right ZipFile instances
    into the lists added, removed, altered, and same.  Returns a tuple
    of added, removed, altered, same
    """

    for event, filename in compare_zips(left, right):
        if event == LEFT:
            group = removed
        elif event == RIGHT:
            group = added
        elif event == DIFF:
            group = altered
        elif event == SAME:
            group = same
        else:
            assert False

        if group is not None:
            group.append(filename)

    return added, removed, altered, same


def is_zipstream(data):
    """
    just like zipfile.is_zipfile, but works upon buffers and streams
    rather than filenames.

    If data supports the read method, it will be treated as a stream
    and read from to test whether it is a valid ZipFile.

    If data also supports the tell and seek methods, it will be
    rewound after being tested.
    """

    if isinstance(data, (str, buffer)):
        data = BytesIO(data)

    if hasattr(data, "read"):
        tell = 0
        if hasattr(data, "tell"):
            tell = data.tell()

        try:
            result = bool(_EndRecData(data))
        except IOError:
            result = False

        if hasattr(data, "seek"):
            data.seek(tell)

    else:
        raise TypeError("requies str, buffer, or stream-like object")

    return result


def file_crc32(filename, chunksize=_CHUNKSIZE):
    """
    calculate the CRC32 of the contents of filename
    """

    check = 0
    with open(filename, 'rb') as fd:
        for data in iter(lambda: fd.read(chunksize), ""):
            check = crc32(data, check)
    return check


def _collect_infos(dirname):

    """ Utility function used by ExplodedZipFile to generate ZipInfo
    entries for all of the files and directories under dirname """

    for r, _ds, fs in walk(dirname):
        if not islink(r) and r != dirname:
            i = ZipInfo()
            i.filename = join(relpath(r, dirname), "")
            i.file_size = 0
            i.compress_size = 0
            i.CRC = 0
            yield i.filename, i

        for f in fs:
            df = join(r, f)
            relfn = relpath(join(r, f), dirname)

            if islink(df):
                pass

            elif isfile(df):
                i = ZipInfo()
                i.filename = relfn
                i.file_size = getsize(df)
                i.compress_size = i.file_size
                i.CRC = file_crc32(df)
                yield i.filename, i

            else:
                # TODO: is there any more special treatment?
                pass


class ExplodedZipFile(object):
    """
    A directory wrapped up to look like a ZipFile. It only populates
    the filename, file_size, and CRC fields of the child ZipInfo
    members.
    """

    def __init__(self, pathname):
        self.fn = pathname
        self.filename = pathname
        self.members = None
        self.refresh()


    def refresh(self):
        self.members = dict(_collect_infos(self.fn))


    def getinfo(self, name):
        return self.members.get(name)


    def namelist(self):
        return sorted(self.members.keys())


    def infolist(self):
        return self.members.values()


    def open(self, name, mode='rb'):
        return open(join(self.fn, name), mode)


    def read(self, name):
        with self.open(name) as fd:
            return fd.read()


    def close(self):
        self.members = None


def zip_file(fn, mode="r"):
    """
    returns either a zipfile.ZipFile instance or an ExplodedZipFile
    instance, depending on whether fn is the name of a valid zip file,
    or a directory.
    """

    if isdir(fn):
        return ExplodedZipFile(fn)
    elif is_zipfile(fn):
        return ZipFile(fn, mode)
    else:
        raise Exception("cannot treat as an archive: %r" % fn)


def open_zip(filename, mode="r"):
    """
    opens a zip file archive at filename in the given mode and returns
    a context manager which will close the archive when the context
    exits. Use eg: with open_zip('my.zip') as z: ...

    In Python 2.6, this will be a ClosingContext instance. In 2.7
    onward, the zipfile.ZipFile class provides its own managed context
    and so the instance is returned unwrapped
    """

    return closing(zip_file(filename, mode))


def open_zip_entry(zipfile, name, mode='r'):
    """
    opens an entry from an opened zip file archive in the given mode
    and returns a context manager which will close the stream when the
    context exits. Use eg: with open_zip_entry(my_zip, 'MANIFEST.MF')
    as data: ...

    In Python 2.6, this will be a ClosingContext instance. In 2.7
    onward, the zipfile.ZipExtFile class provides its own managed
    context and so the instance is returned unwrapped
    """

    return closing(zipfile.open(name, mode))


def chunk_zip_entry(zipfile, name, chunksize=_CHUNKSIZE):
    """
    opens an entry from an openex zip file archive and yields
    sequential chunks of data from the resulting stream.
    """

    with open_zip_entry(zipfile, name, mode='r') as stream:
        data = stream.read(chunksize)
        while data:
            yield data
            data = stream.read(chunksize)


def zip_entry_rollup(zipfile):
    """
    returns a tuple of (files, dirs, size_uncompressed,
    size_compressed). files+dirs will equal len(zipfile.infolist)
    """

    files = dirs = 0
    total_c = total_u = 0

    for i in zipfile.infolist():
        if i.filename[-1] == '/':
            # I wonder if there's a better detection method than this
            dirs += 1
        else:
            files += 1
            total_c += i.compress_size
            total_u += i.file_size

    return files, dirs, total_c, total_u


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Replacement setuptools/distutils build_py command which also
compiles Cheetah templates to Python

author: Christopher O'Brien  <obriencj@gmail.com>
license: LGPL v.3
"""


from distutils.core import Command
from distutils.util import newer
from distutils.command.build_py import build_py
from glob import glob
from os import makedirs
from os.path import basename, exists, join, splitext


DEFAULT_CONFIG = "extras/cheetah.cfg"


class cheetah_cmd(Command):
    """
    command that compiles Cheetah template files into Python
    """

    # TODO: move the act of actually compiling .tmpl -> .py into this
    # command, and simply make the build_py command depend on this one

    pass


class cheetah_build_py_cmd(build_py):
    """
    build_py command with some special handling for Cheetah template
    files. Takes tmpl from package source directories and compiles
    them for distribution. This allows me to write tmpl files in the
    src dir of my project, and have them get compiled to py/pyc/pyo
    files during the build process.
    """

    # Note: it's important to override build_py rather than to add a
    # sub-command to build. The build command doesn't collate the
    # get_outputs of its sub-commands, and install specifically looks
    # for build_py and build_ext for the list of files to install.

    # Update: now that I've switched from distutils to setuptools, it
    # may be possible to put this into a subcommand instead. Have to
    # investigate.

    def find_package_templates(self, package, package_dir):
        # template files will be located under src, and will end in .tmpl

        self.check_package(package, package_dir)
        template_files = glob(join(package_dir, "*.tmpl"))
        templates = []

        for f in template_files:
            template = splitext(basename(f))[0]
            templates.append((package, template, f))
        return templates


    def build_package_templates(self):
        for package in self.packages:
            package_dir = self.get_package_dir(package)
            templates = self.find_package_templates(package, package_dir)

            for package_, template, template_file in templates:
                assert package == package_
                self.build_template(template, template_file, package)


    def build_template(self, template, template_file, package):
        """
        Compile the cheetah template in src into a python file in build
        """

        try:
            from Cheetah.Compiler import Compiler
        except ImportError:
            self.announce("unable to import Cheetah.Compiler, build failed")
            raise
        else:
            comp = Compiler(file=template_file, moduleName=template)

        # load configuration if it exists
        conf_fn = DEFAULT_CONFIG
        if exists(conf_fn):
            with open(conf_fn, "rt") as config:
                comp.updateSettingsFromConfigFileObj(config)

        # and just why can't I configure these?
        comp.setShBang("")
        comp.addModuleHeader("pylint: disable=C,W,R,F")

        outfd = join(self.build_lib, *package.split("."))
        outfn = join(outfd, template + ".py")

        if not exists(outfd):
            makedirs(outfd)

        if newer(template_file, outfn):
            self.announce("compiling %s -> %s" % (template_file, outfd), 2)
            with open(outfn, "w") as output:
                output.write(str(comp))


    def get_template_outputs(self, include_bytecode=1):
        built = list()

        for package in self.packages:
            package_dir = self.get_package_dir(package)
            templates = self.find_package_templates(package, package_dir)
            for _, template, _ in templates:
                outfd = join(self.build_lib, *package.split("."))
                outfn = join(outfd, template + ".py")

                built.append(outfn)

                if include_bytecode:
                    if self.compile:
                        built.append(outfn + "c")
                    if self.optimize > 0:
                        built.append(outfn + "o")

        return built


    def get_outputs(self, include_bytecode=1):
        # Overridden to append our compiled templates in addition to
        # the normal build outputs.

        outputs = build_py.get_outputs(self, include_bytecode)
        outputs.extend(self.get_template_outputs(include_bytecode))

        return outputs


    def run(self):
        if self.packages:
            self.build_package_templates()

        # old-style class, can't use super
        build_py.run(self)


#
# The end.



# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see
# <http://www.gnu.org/licenses/>.


"""
Module for reading and writing files, related to JAR manifest.

References
----------
* http://docs.oracle.com/javase/1.5.0/docs/guide/jar/index.html
* http://java.sun.com/j2se/1.5.0/docs/guide/jar/jar.html#JAR%20Manifest

:author: Christopher O'Brien  <obriencj@gmail.com>
:license: LGPL
"""

from __future__ import print_function

import argparse
import hashlib
import os
import sys

from base64 import b64encode
from collections import OrderedDict
from os.path import isdir, join, sep, split
from os import walk
from six import BytesIO
from six.moves import zip
from zipfile import ZipFile

from .change import GenericChange, SuperChange
from .change import Addition, Removal
from .dirutils import fnmatches, makedirsp

try:
    buffer
except NameError:
    buffer = memoryview

__all__ = (
    "ManifestChange", "ManifestSectionChange",
    "ManifestSectionAdded", "ManifestSectionRemoved",
    "Manifest", "ManifestSection",
    "SignatureManifest",
    "ManifestKeyException", "MalformedManifest",
    "main", "cli_create", "cli_query", "cli_verify", )


_BUFFERING = 2 ** 14


SIG_FILE_PATTERN = "*.SF"
SIG_BLOCK_PATTERNS = ("*.RSA", "*.DSA", "*.EC", "SIG-*", )


class UnsupportedDigest(Exception):
    """
    Indicates an algorithm was requested for which we had no matching
    digest support
    """
    pass


class ManifestKeyException(Exception):
    """
    Indicates there was an issue with the key used in a manifest
    section
    """
    pass


class MalformedManifest(Exception):
    """
    Indicates there was a problem in parsing a manifest
    """
    pass


# Digests classes by Java name that have been found present in hashlib
JAVA_TO_OPENSSL_DIGESTS = {}
JAVA_TO_OPENSSL_DIGEST_NAMES = {}


def _add_digest(java_name, hashlib_name):
    digest = getattr(hashlib, hashlib_name, None)
    if digest:
        JAVA_TO_OPENSSL_DIGESTS[java_name] = digest
        JAVA_TO_OPENSSL_DIGEST_NAMES[java_name] = hashlib_name


def _get_digest(java_name, as_string=False):
    try:
        return JAVA_TO_OPENSSL_DIGEST_NAMES[java_name] if as_string \
            else JAVA_TO_OPENSSL_DIGESTS[java_name]
    except KeyError:
        raise UnsupportedDigest("Unsupported digest %s. Supported: %s" %
                                (java_name, ", ".join(sorted(
                                 JAVA_TO_OPENSSL_DIGESTS.keys()))))


# Note 1: Java supports also MD2, but hashlib does not
_add_digest("MD2", "md2")
_add_digest("MD5", "md5")

# Note 2: Oracle specifies "SHA-1" algorithm name in their
# documentation, but it's called "SHA1" elsewhere and that is what
# jarsigner uses as well.
# http://docs.oracle.com/javase/8/docs/technotes/guides/security/StandardNames.html#MessageDigest
_add_digest("SHA1", "sha1")
_add_digest("SHA-256", "sha256")
_add_digest("SHA-384", "sha384")
_add_digest("SHA-512", "sha512")


class ManifestSectionChange(GenericChange):
    label = "Manifest Subsection"


    def get_description(self):
        m = self.ldata or self.rdata
        entry = m.primary()
        if self.is_change():
            return "%s Changed: %s" % (self.label, entry)
        else:
            return "%s Unchanged: %s" % (self.label, entry)


    def is_ignored(self, options):
        if getattr(options, "ignore_manifest_subsections", False):
            return True

        ikeys = set(getattr(options, "ignore_manifest_key", set()))
        if ikeys:
            lset = set(self.ldata.items())
            rset = set(self.rdata.items())
            changed = set(k for k, _v in lset.symmetric_difference(rset))
            return changed.issubset(ikeys)

        else:
            return False


class ManifestSectionAdded(ManifestSectionChange, Addition):

    label = "Manifest Subsection Added"

    def get_description(self):
        return "%s: %s" % (self.label, self.rdata.primary())


    def is_ignored(self, options):
        return getattr(options, "ignore_manifest_subsections", False)


class ManifestSectionRemoved(ManifestSectionChange, Removal):

    label = "Manifest Subsection Removed"

    def get_description(self):
        return "%s: %s" % (self.label, self.ldata.primary())


    def is_ignored(self, options):
        return getattr(options, "ignore_manifest_subsections", False)


class ManifestMainChange(GenericChange):

    label = "Manifest Main Section"


    def get_description(self):
        if self.is_change():
            return "%s has changed" % self.label
        else:
            return "%s is unchanged" % self.label


    def is_ignored(self, options):
        ikeys = set(getattr(options, "ignore_manifest_key", set()))
        if ikeys:
            lset = set(self.ldata.items())
            rset = set(self.rdata.items())
            changed = set(k for k, _v in lset.symmetric_difference(rset))
            return changed.issubset(ikeys)

        else:
            return False


class ManifestChange(SuperChange):

    label = "Manifest"


    def collect_impl(self):
        lm, rm = self.ldata, self.rdata
        yield ManifestMainChange(lm, rm)

        l_sections = set(lm.sub_sections.keys())
        r_sections = set(rm.sub_sections.keys())

        for s in l_sections.intersection(r_sections):
            yield ManifestSectionChange(lm.sub_sections[s], rm.sub_sections[s])

        for s in l_sections.difference(r_sections):
            yield ManifestSectionRemoved(lm.sub_sections[s], None)

        for s in r_sections.difference(l_sections):
            yield ManifestSectionAdded(None, rm.sub_sections[s])


    def is_ignored(self, options):
        return getattr(options, "ignore_manifest", False) or \
            SuperChange.is_ignored(self, options)


class ManifestSection(OrderedDict):

    primary_key = "Name"


    def __init__(self, name=None):
        OrderedDict.__init__(self)
        self[self.primary_key] = name


    def __gt__(self, other_section):
        # we need just some ordering, no matter which.
        return self.get_data() > other_section.get_data()

    def __setitem__(self, k, v):
        # pylint: disable=W0221
        # we want the behavior of OrderedDict, but don't take the
        # additional parameter

        # our keys should always be strings, as should our values. We
        # also have an upper limit on the length we can permit for
        # keys, per the JAR MANIFEST specification.
        k = str(k)
        if len(k) > 68:
            raise ManifestKeyException("key too long", k)
        else:
            OrderedDict.__setitem__(self, k, str(v))


    def primary(self):
        """
        The primary value for this section
        """

        return self.get(self.primary_key)


    def load(self, items):
        """
        Populate this section from an iteration of the parse_items call
        """

        for k, vals in items:
            self[k] = "".join(vals)


    def store(self, stream, linesep=os.linesep):
        """
        Serialize this section and write it to a binary stream
        """

        for k, v in self.items():
            write_key_val(stream, k, v, linesep)

        stream.write(linesep.encode('utf-8'))


    def get_data(self, linesep=os.linesep):
        """
        Serialize the section and return it as bytes
        :return bytes
        """

        stream = BytesIO()
        self.store(stream, linesep)
        return stream.getvalue()


    def keys_with_suffix(self, suffix):
        """
        :return: list of keys ending with given :suffix:.
        """
        return [k.rstrip(suffix) for k in self.keys() if k.endswith(suffix)]


class Manifest(ManifestSection):
    """
    Represents a Java Manifest as an ordered dictionary containing
    the key:value pairs from the main section of the manifest, and
    zero or more sub-dictionaries of key:value pairs representing the
    sections following the main section. The sections are referenced
    by the value of their 'Name' pair, which must be unique to the
    Manifest as a whole.
    """

    primary_key = "Manifest-Version"


    def __init__(self, version="1.0", linesep=None):
        # can't use super, because we're a child of a non-object
        ManifestSection.__init__(self, version)
        self.sub_sections = OrderedDict([])
        self.linesep = linesep


    def create_section(self, name, overwrite=True):
        """
        create and return a new sub-section of this manifest, with the
        given Name attribute. If a sub-section already exists with
        that name, it will be lost unless overwrite is False in which
        case the existing sub-section will be returned.
        """

        if overwrite:
            sect = ManifestSection(name)
            self.sub_sections[name] = sect

        else:
            sect = self.sub_sections.get(name, None)
            if sect is None:
                sect = ManifestSection(name)
                self.sub_sections[name] = sect

        return sect


    def parse_file(self, filename):
        """
        Load self from the file, such as "MANIFEST.MF" or "SIGNATURE.SF".
        :param filename: contains UTF-8 encoded manifest
        """

        with open(filename, "rb", _BUFFERING) as stream:
            self.parse(stream.read())


    def load_from_jar(self, jarfile):
        # Can't be imported at top level:
        from javatools.jarutil import MissingManifestError
        with ZipFile(jarfile) as jar:
            if "META-INF/MANIFEST.MF" not in jar.namelist():
                raise MissingManifestError(
                    "META-INF/MANIFEST.MF not found in %s" % jarfile)
            data = jar.read("META-INF/MANIFEST.MF")
            self.parse(data)


    def parse(self, data):
        """
        populate instance with values and sub-sections
        :param data: UTF-8 encoded manifest
        :type data: bytes
        """

        # we want at least some data, thus we ignore unknown characters
        data = data.decode('utf-8', errors="ignore")
        self.linesep = detect_linesep(data)

        # the first section is the main one for the manifest. It's
        # also where we will check for our newline separator
        sections = parse_sections(data)
        self.load(next(sections))

        # and all following sections are considered sub-sections
        for section in sections:
            next_section = ManifestSection(None)
            next_section.load(section)
            self.sub_sections[next_section.primary()] = next_section


    def store(self, stream, linesep=None):
        """
        Serialize the Manifest to a binary stream
        """

        # either specified here, specified on the instance, or the OS
        # default
        linesep = linesep or self.linesep or os.linesep

        ManifestSection.store(self, stream, linesep)
        for sect in sorted(self.sub_sections.values()):
            sect.store(stream, linesep)


    def get_main_section(self, linesep=None):
        """
        Serialize just the main section of the manifest and return it as bytes
        :return bytes
        """

        linesep = linesep or self.linesep or os.linesep

        stream = BytesIO()
        ManifestSection.store(self, stream, linesep)
        return stream.getvalue()


    def get_data(self, linesep=None):
        """
        Serialize the entire manifest and return it as bytes
        :return bytes
        """

        linesep = linesep or self.linesep or os.linesep

        stream = BytesIO()
        self.store(stream, linesep)
        return stream.getvalue()


    def verify_jar_checksums(self, jar_file, strict=True):
        """
        Verify checksums, present in the manifest, against the JAR content.
        :return: list of entries for which verification has failed
        """

        verify_failures = []

        zip_file = ZipFile(jar_file)
        for filename in zip_file.namelist():
            if file_skips_verification(filename):
                continue

            file_section = self.create_section(filename, overwrite=False)

            digests = file_section.keys_with_suffix("-Digest")
            if not digests and strict:
                verify_failures.append(filename)
                continue

            for java_digest in digests:
                read_digest = file_section.get(java_digest + "-Digest")
                calculated_digest = b64_encoded_digest(
                    zip_file.read(filename),
                    _get_digest(java_digest))

                if calculated_digest == read_digest:
                    # found a match
                    break
            else:
                # for all the digests, not one of them matched. Add
                # this filename to the error list
                verify_failures.append(filename)

        return verify_failures


    def add_jar_entries(self, jar_file, digest_name="SHA-256"):
        """
        Add manifest sections for all but signature-related entries
        of :param jar_file.
        :param digest_name The digest algorithm to use
        :return None
        TODO: join code with cli_create
        """

        key_digest = digest_name + "-Digest"
        digest = _get_digest(digest_name)

        with ZipFile(jar_file, 'r') as jar:
            for entry in jar.namelist():
                if file_skips_verification(entry):
                    continue
                section = self.create_section(entry)
                section[key_digest] = b64_encoded_digest(jar.read(entry),
                                                         digest)


    def clear(self):
        """
        removes all items from this manifest, and clears and removes all
        sub-sections
        """

        for sub in self.sub_sections.values():
            sub.clear()
        self.sub_sections.clear()

        ManifestSection.clear(self)


    def __del__(self):
        self.clear()


class SignatureManifest(Manifest):
    """
    Represents a KEY.SF signature file.  Structure is similar to that
    of Manifest. Each section represents a crypto checksum of a matching
    section from a MANIFEST.MF
    """

    primary_key = "Signature-Version"


    def digest_manifest(self, manifest, java_algorithm="SHA-256"):
        """
        Create a main section checksum and sub-section checksums based off
        of the data from an existing manifest using an algorithm given
        by Java-style name.
        """

        # pick a line separator for creating checksums of the manifest
        # contents. We want to use either the one from the given
        # manifest, or the OS default if it hasn't specified one.
        linesep = manifest.linesep or os.linesep

        all_key = java_algorithm + "-Digest-Manifest"
        main_key = java_algorithm + "-Digest-Manifest-Main-Attributes"
        sect_key = java_algorithm + "-Digest"

        digest = _get_digest(java_algorithm)
        accum = manifest.get_main_section()
        self[main_key] = b64_encoded_digest(accum, digest)

        for sub_section in manifest.sub_sections.values():
            sub_data = sub_section.get_data(linesep)
            sf_sect = self.create_section(sub_section.primary())
            sf_sect[sect_key] = b64_encoded_digest(sub_data, digest)
            accum += sub_data

        self[all_key] = b64_encoded_digest(accum, digest)

    def verify_manifest_main_checksum(self, manifest):
        """
        Verify the checksum over the manifest main section.

        :return: True if the signature over main section verifies
        """

        # NOTE: JAR spec does not state whether there can be >1 digest used,
        # and should the validator require any or all digests to match.
        # We allow mismatching digests and require just one to be correct.
        # We see no security risk: it is the signer of the .SF file
        # who shall check, what is being signed.
        for java_digest in self.keys_with_suffix("-Digest-Manifest"):
            whole_mf_digest = b64_encoded_digest(
                manifest.get_data(),
                _get_digest(java_digest))

            # It is enough for at least one digest to be correct
            if whole_mf_digest == self.get(java_digest + "-Digest-Manifest"):
                return True

        return False


    def verify_manifest_main_attributes_checksum(self, manifest):
        # JAR spec allows for the checksum of the whole manifest to mismatch.
        # There is a second chance for the verification to succeed:
        # checksum for the main section matches,
        # plus checksum for every subsection matches.

        keys = self.keys_with_suffix("-Digest-Manifest-Main-Attributes")
        for java_digest in keys:
            mf_main_attr_digest = b64_encoded_digest(
                manifest.get_main_section(),
                _get_digest(java_digest))

            attr = java_digest + "-Digest-Manifest-Main-Attributes"
            if mf_main_attr_digest == self.get(attr):
                return True
        else:
            return False


    def verify_manifest_entry_checksums(self, manifest, strict=True):
        """
        Verifies the checksums over the given manifest. If strict is True
        then entries which had no digests will fail verification. If
        strict is False then entries with no digests will not be
        considered failing.

        :return: List of entries which failed to verify.

        Reference:
        http://docs.oracle.com/javase/7/docs/technotes/guides/jar/jar.html#Signature_Validation
        """

        # TODO: this behavior is probably wrong -- surely if there are
        # multiple digests, they should ALL match, or verification would
        # fail?

        failures = []

        for s in manifest.sub_sections.values():
            sf_section = self.create_section(s.primary(), overwrite=False)

            digests = s.keys_with_suffix("-Digest")
            if not digests and strict:
                failures.append(s.primary())
                continue

            for java_digest in digests:
                section_digest = b64_encoded_digest(
                    s.get_data(manifest.linesep),
                    _get_digest(java_digest))

                if section_digest == sf_section.get(java_digest + "-Digest"):
                    # found a match, verified
                    break
            else:
                # no matches found for the digests present
                failures.append(s.primary())

        return failures


    def verify_manifest(self, manifest):
        if self.verify_manifest_main_checksum(manifest):
            return []
        # Main checksum does not validate, but there is a second chance.
        if not self.verify_manifest_main_attributes_checksum(manifest):
            # Let's return the manifest itself - such return value can be
            # interpreted only in exactly this way
            return ["META-INF/MANIFEST.MF"]

        return self.verify_manifest_entry_checksums(manifest)


    def get_signature(self, certificate, private_key, extra_certs,
                      digest_algorithm="SHA-256"):

        from .crypto import create_signature_block

        openssl_digest = _get_digest(digest_algorithm, as_string=True)
        return create_signature_block(openssl_digest, certificate, private_key,
                                      extra_certs, self.get_data())


class SignatureManifestChange(ManifestChange):
    label = "Signature File"

    def is_ignored(self, options):
        return getattr(options, "ignore_jar_signature", False) or \
            SuperChange.is_ignored(self, options)


class SignatureBlockFileChange(GenericChange):
    label = "Signature Block File"

    def get_description(self):
        return "[binary data change]"

    def fn_pretty(self, side_data):
        return "[binary data]"


def _b64encode_to_str(data):
    """
    Wrapper around b64encode which takes and returns same-named types
    on both Python 2 and Python 3.
    :type data: bytes
    :return: str
    """
    ret = b64encode(data)
    if not isinstance(ret, str):  # Python3
        return ret.decode('ascii')
    else:
        return ret


def b64_encoded_digest(data, algorithm):
    """
    :type data: bytes
    :return: str
    """
    h = algorithm()
    h.update(data)
    return _b64encode_to_str(h.digest())


def detect_linesep(data):
    """
    :type data: unicode in Py2, str in Py3
    """
    return "\r\n" if "\r\n" in data else "\n"


def parse_sections(data):
    """
    yields one section at a time in the form

    [ (key, [val, ...]), ... ]

    where key is a string and val is a string representing a single
    line of any value associated with the key. Multiple vals may be
    present if the value is long enough to require line continuations
    in which case simply concatenate the vals together to get the full
    value.
    :type data: unicode in Py2, str in Py3
    """

    if not data:
        return

    # our current section
    curr = None

    for lineno, line in enumerate(data.splitlines()):
        # Clean up the line
        cleanline = line.replace('\x00', '')

        if not cleanline:
            # blank line means end of current section (if any)
            if curr:
                yield curr
                curr = None

        elif cleanline[0] == ' ':
            # line beginning with a space means a continuation
            if curr is None:
                raise MalformedManifest("bad line continuation, "
                                        " line: %i" % lineno)
            else:
                # pylint: disable=unsubscriptable-object
                curr[-1][1].append(cleanline[1:])

        else:
            # otherwise the beginning of a new k:v pair
            if curr is None:
                curr = list()

            try:
                key, val = cleanline.split(':', 1)
                curr.append((key, [val[1:]]))
            except ValueError:
                raise MalformedManifest(
                    "Invalid manifest line: %i; line contents: %s"
                    % (lineno, cleanline))


    # yield and leftovers
    if curr:
        yield curr


def write_key_val(stream, key, val, linesep=os.linesep):
    """
    The MANIFEST specification limits the width of individual lines to
    72 bytes (including the terminating newlines). Any key and value
    pair that would be longer must be split up over multiple
    continuing lines
    :type key, val: str in Py3, str or unicode in Py2
    :type stream: binary
    """

    key = key.encode('utf-8') or ""
    val = val.encode('utf-8') or ""
    linesep = linesep.encode('utf-8')

    if not (0 < len(key) < 69):
        raise ManifestKeyException("bad key length", key)

    if len(key) + len(val) > 68:
        kvbuffer = BytesIO(b": ".join((key, val)))

        # first grab 70 (which is 72 after the trailing newline)
        stream.write(kvbuffer.read(70))

        # now only 69 at a time, because we need a leading space and a
        # trailing \n
        part = kvbuffer.read(69)
        while part:
            stream.write(linesep + b" ")
            stream.write(part)
            part = kvbuffer.read(69)
        kvbuffer.close()

    else:
        stream.write(key)
        stream.write(b": ")
        stream.write(val)

    stream.write(linesep)


def digest_chunks(chunks, algorithms=(hashlib.md5, hashlib.sha1)):
    """
    returns a base64 rep of the given digest algorithms from the
    chunks of data
    """

    hashes = [algorithm() for algorithm in algorithms]

    for chunk in chunks:
        for h in hashes:
            h.update(chunk)

    return [_b64encode_to_str(h.digest()) for h in hashes]


def file_chunk(filename, size=_BUFFERING):
    """
    returns a generator function which when called will emit x-sized
    chunks of filename's contents
    """

    def chunks():
        with open(filename, "rb", _BUFFERING) as fd:
            buf = fd.read(size)
            while buf:
                yield buf
                buf = fd.read(size)
    return chunks


def zipentry_chunk(zipfile, name, size=_BUFFERING):
    """
    returns a generator function which when called will emit x-sized
    chunks of the named entry in the zipfile object
    """

    def chunks():
        with zipfile.open(name) as fd:
            buf = fd.read(size)
            while buf:
                yield buf
                buf = fd.read(size)
    return chunks


def directory_generator(dirname, trim=0):
    """
    yields a tuple of (relative filename, chunking function). The
    chunking function can be called to open and iterate over the
    contents of the filename.
    """

    def gather(collect, dirname, fnames):
        for fname in fnames:
            df = join(dirname, fname)
            if not isdir(df):
                collect.append(df)

    collect = list()
    walk(dirname, gather, collect)
    for fname in collect:
        yield fname[trim:], file_chunk(fname)


def multi_path_generator(pathnames):
    """
    yields (name,chunkgen) for all of the files found under the list
    of pathnames given. This is recursive, so directories will have
    their contents emitted. chunkgen is a function that can called and
    iterated over to obtain the contents of the file in multiple
    reads.
    """

    for pathname in pathnames:
        if isdir(pathname):
            for entry in directory_generator(pathname):
                yield entry
        else:
            yield pathname, file_chunk(pathname)


def _in_sig_related_dir(filename):
    path_components = filename.split("/")
    return len(path_components) == 2 and path_components[0] == "META-INF"


def file_matches_sigfile(filename):
    return _in_sig_related_dir(filename) \
        and fnmatches(filename.upper(), SIG_FILE_PATTERN)


def file_matches_sigblock(filename):
    return _in_sig_related_dir(filename) \
        and fnmatches(filename[len("META-INF/"):].upper(), *SIG_BLOCK_PATTERNS)


def file_skips_verification(filename):
    # http://docs.oracle.com/javase/8/docs/technotes/guides/jar/jar.html#SignedJar-Overview

    filename = filename.upper()
    return filename == "META-INF/MANIFEST.MF" \
        or filename.endswith("/") \
        or file_matches_sigblock(filename) \
        or file_matches_sigfile(filename)


def single_path_generator(pathname):
    """
    emits name,chunkgen pairs for the given file at pathname. If
    pathname is a directory, will act recursively and will emit for
    each file in the directory tree chunkgen is a generator that can
    be iterated over to obtain the contents of the file in multiple
    parts
    """

    if isdir(pathname):
        trim = len(pathname)
        if pathname[-1] != sep:
            trim += 1
        for entry in directory_generator(pathname, trim):
            yield entry

    else:
        zf = ZipFile(pathname)
        for f in zf.namelist():
            if f[-1] != '/':
                yield f, zipentry_chunk(zf, f)
        zf.close()


def cli_create(argument_list):
    """
    command-line call to create a manifest from a JAR file or a
    directory
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("content", help="file or directory")
    # TODO: shouldn't we always process directories recursively?
    parser.add_argument("-r", "--recursive",
                        help="process directories recursively")
    parser.add_argument("-i", "--ignore", nargs="+", action="append",
                        help="patterns to ignore "
                             "(can be given more than once)")
    parser.add_argument("-m", "--manifest", default=None,
                        help="output file (default is stdout)")
    parser.add_argument("-d", "--digest",
                        help="digest(s) to use, comma-separated")

    args = parser.parse_args(argument_list)

    # TODO: remove digest from here, they are created when signing!
    if args.digest is None:
        args.digest = "MD5,SHA1"
    requested_digests = args.digest.split(",")
    use_digests = [_get_digest(digest) for digest in requested_digests]

    if args.recursive:
        entries = multi_path_generator(args.content)
    else:
        entries = single_path_generator(args.content)

    mf = Manifest()

    ignores = ["META-INF/*"]
    if args.ignore:
        ignores.extend(*args.ignore)

    for name, chunks in entries:
        # skip the stuff that we were told to ignore
        if ignores and fnmatches(name, *ignores):
            continue

        sec = mf.create_section(name)

        digests = zip(requested_digests, digest_chunks(chunks(), use_digests))
        for digest_name, digest_value in digests:
            sec[digest_name + "-Digest"] = digest_value

    if args.manifest:
        # we'll output to the manifest file if specified, and we'll
        # even create parent directories for it, if necessary
        makedirsp(split(args.manifest)[0])
        output = open(args.manifest, "wb")
    else:
        output = sys.stdout

    mf.store(output)

    if args.manifest:
        output.close()


def cli_verify(args):
    if len(args) != 1 or "-h" in args:
        print("Usage: manifest v [--ignore=PATH] file.jar")
        return 2

    jarfn = args[0]
    mf = Manifest()
    mf.load_from_jar(jarfn)

    errors = mf.verify_jar_checksums(jarfn)
    if len(errors) > 0:
        print("Verify failed, no matching checksums for files: %s"
              % ", ".join(errors))
        return 1

    else:
        return 0


def cli_query(args):
    if len(args) < 2 or "-h" in args:
        print("Usage: manifest q file.jar key_to_query...")
        return 1

    mf = Manifest()
    mf.load_from_jar(args[0])

    for q in args[1:]:
        s = q.split(':', 1)
        if len(s) > 1:
            mfs = mf.sub_sections.get(s[0])
            if mfs:
                print(q, "=", mfs.get(s[1]))
            else:
                print(q, ": No such section")

        else:
            print(q, "=", mf.get(s[0]))


def usage(error_msg=None):
    if error_msg is not None:
        print(error_msg)
    print("Usage: manifest <command> [options] <argument>")
    print("Commands:")
    print("    c: create a manifest")
    print("    q: query manifest for values")
    print("    v: verify manifest checksums")
    print("Give option \"-h\" for help on a particular command.")
    return 1


def main(args=sys.argv):
    """
    main entry point for the manifest CLI
    """

    if len(args) < 2:
        return usage("Command expected")

    command = args[1]
    rest = args[2:]

    if "create".startswith(command):
        return cli_create(rest)
    elif "query".startswith(command):
        return cli_query(rest)
    elif "verify".startswith(command):
        return cli_verify(rest)
    else:
        return usage("Unknown command: %s" % command)

#
# The end.
